package net.pricefx.common.dto;

import java.util.*;
import java.util.stream.Collectors;

import net.pricefx.common.api.WorkflowType;

/**
 * This class is basically the type of object, which is exposed in the workflow logic by
 * the binding variable {@code workflow}
 * <p>In your workflow logic you would call methods on that class instance to add
 * approval nodes and/or watcher nodes and supply further details
 * <p>The methods are designed to be used in a simple flow pattern.
 * <p>Simple example:
 * <blockquote><pre>
 * workflow.addApprovalStep("First Approval").withApprovers("admin").withReasons("Admin needs always approve")
 * </pre></blockquote>
 */
public class WorkflowDTO {

	private WorkflowType type;
	private String typedId;
	private Map<String, String> itemReasons;
	private final LinkedHashSet<StepDTO> steps = new LinkedHashSet<>();
	private List<String> denialReasons = new ArrayList<>();
	private boolean isDenialReasonOptional = true;
	private boolean isAddStepRestrictedToWFAdmin = false;
	private boolean isSelectStepOnDeny = false;
	private Map<String, Object> dataMap = new LinkedHashMap<>();
	private String defaultPostApprovalStepLogicName;
	private String submitReason;
	private boolean runDefaultPostStepLogicOnEmptyWorkflow;
	private Boolean autoConvertToDeal;
	private boolean isMassUpdate;

	public WorkflowDTO(WorkflowType type, String typedId, String submitReason, boolean isMassUpdate) {
		this(type, typedId);
		this.isMassUpdate = isMassUpdate;
		this.submitReason = submitReason;
	}

	public WorkflowDTO(WorkflowType type, String typedId) {
		this.type = type;
		this.typedId = typedId;
	}

	/**
	 * Adds a new approval step at the end of the approval workflow.
	 *
	 * @param uniqueName Name of the step, it must be unique inside one approval workflow.
	 *                      If you will add again a Step with the same unique name, you will override the existing one.
	 * @return The Approval Step itself.
	 */
	public ApprovalStepDTO addApprovalStep(String uniqueName) {
		ApprovalStepDTO qas = new ApprovalStepDTO(this, uniqueName);
		steps.add(qas);
		return qas;
	}

	/**
	 * Adds a watcher step to the end of the approval workflow.
	 * A watcher step cannot approve or deny, but will receive a notification when it becomes active.
	 * @param uniqueName Name of the step. It must be unique inside one approval flow.
	 *                      If you will add again a Step with the same unique name, you will override the existing one.
	 * @return The Watcher Step itself.
	 */
	public WatcherStepDTO addWatcherStep(String uniqueName) {
		WatcherStepDTO ws = new WatcherStepDTO(this, uniqueName);
		steps.add(ws);
		return ws;
	}

	/**
	 * Returns the type of the workflow. Corresponds to the underlying approvable document:
	 * <ul>
	 *     <li>QUOTE("quote")</li>
	 *     <li>PRICEGRIDITEM("pricegriditem")</li>
	 *     <li>REBATEAGREEMENT("rebateagreement")</li>
	 *     <li>CONTRACT("contract")</li>
	 *     <li>DATACHANGEREQUEST("dcr")</li>
	 *     <li>PRICELIST("pricelist")</li>
	 *     <li>REBATERECORD("rebaterecord")</li>
	 *     <li>MODELRECORD("modelrecord")</li>
	 * 	   <li>CLAIM("claim")</li>
	 * </ul>
	 *
	 * @return The Type of the Workflow.
	 */
	public WorkflowType getType() {
		return type;
	}

	/**
	 * Removes a workflow step with given unique name (if it exists).
	 * @param uniqueName The name of the step
	 * @return
	 */
	public boolean removeStep(String uniqueName) {
		return steps.remove(new ApprovalStepDTO(this, uniqueName));
	}

	/**
	 * Retrieves a workflow step with given name (if it exists)
	 * @param uniqueName The name of the step
	 * @return
	 */
	public StepDTO getStep(final String uniqueName) {
		for (StepDTO stepDTO : steps) {
			if (uniqueName.equals(stepDTO.uniqueName)) {
				return stepDTO;
			}
		}
		return null;
	}

	/**
	 * Gets reason entered by submitter
	 * @return
	 */
	public String getSubmitReason() { return submitReason; }

	/**
	 * Overrides reason entered by submitter
	 * @param submitReason
	 */
	public void setSubmitReason(String submitReason) {
		this.submitReason = submitReason;
	}

	public void addItemReason(String itemTypedId, String reason) {
		if (itemReasons == null)
			itemReasons = new HashMap<>();
		itemReasons.put(itemTypedId, reason);
	}

	public Map<String, String> getItemReasons() {
		return itemReasons;
	}

	public WorkflowDTO withDenialReasons(List<String> denialReasons, boolean isDenialReasonOptional) {
		this.denialReasons.addAll(denialReasons);
		this.isDenialReasonOptional = isDenialReasonOptional;
		return this;
	}

	public List<String> getDenialReasons() {
		return denialReasons;
	}

	public boolean isDenialReasonOptional() {
		return isDenialReasonOptional;
	}

	// WorkflowDTO framework methods, not intented to be called by user
	public String getTypedId() {
		return typedId;
	}

	public LinkedHashSet<StepDTO> getSteps() {
		return steps;
	}

	public WorkflowDTO withAddStepRestrictedToWFAdmin() {
		this.isAddStepRestrictedToWFAdmin = true;
		return this;
	}

	public WorkflowDTO withAddStepRestrictedToWFAdmin(boolean isAddStepRestrictedToWFAdmin) {
		this.isAddStepRestrictedToWFAdmin = isAddStepRestrictedToWFAdmin;
		return this;
	}

	public boolean isAddStepRestrictedToWFAdmin() {
		return isAddStepRestrictedToWFAdmin;
	}

	public WorkflowDTO withSelectStepOnDeny(boolean isSelectStepOnDeny) {
		this.isSelectStepOnDeny = isSelectStepOnDeny;
		return this;
	}

	public boolean isSelectStepOnDeny() {
		return isSelectStepOnDeny;
	}

	public Map<String, Object> getDataMap() {
		return dataMap;
	}

	/**
	 * Sets dataMap map, which is available then in email templates
	 * @param dataMap The map with additional data needed to render the template.
	 * @return
	 */
	public WorkflowDTO withDataMap(Map<String, Object> dataMap) {
		this.dataMap = dataMap;
		return this;
	}

	/**
	 * Adds a logic that will be executed after every approval step.
	 *
	 * @param postApprovalStepLogicUniqueName Unique name of wf post step logic
	 */
	public WorkflowDTO withDefaultPostApprovalStepLogic(String postApprovalStepLogicUniqueName) {
		this.defaultPostApprovalStepLogicName = postApprovalStepLogicUniqueName;
		return this;
	}

	public String getDefaultPostApprovalStepLogicName() {
		return defaultPostApprovalStepLogicName;
	}

	/**
	 * Decides if default post step logic should be executed also in case the wf has no approval steps (i.e. empty or all steps are watcher steps).
	 *
	 * @param runDefaultPostStepLogicOnEmptyWorkflow Default is false.
	 */
	public WorkflowDTO withRunDefaultPostApprovalStepLogicOnEmptyWorkflow(boolean runDefaultPostStepLogicOnEmptyWorkflow) {
		this.runDefaultPostStepLogicOnEmptyWorkflow = runDefaultPostStepLogicOnEmptyWorkflow;
		return this;
	}

	public boolean getRunDefaultPostApprovalStepLogicOnEmptyWorkflow() {
		return runDefaultPostStepLogicOnEmptyWorkflow;
	}

	/**
	 * If <code>true</code> then automatically converts quote to deal when quote is approved
	 *
	 * @param autoConvertToDeal flag if should be automatically converted
	 */
	public WorkflowDTO withAutoConvertToDeal(boolean autoConvertToDeal) {
		this.autoConvertToDeal = autoConvertToDeal;
		return this;
	}

	public Boolean isAutoConvertToDeal() {
		return autoConvertToDeal;
	}

	/**
	 * @return true If workflow is called from mass update job
	 */
	public boolean isMassUpdate() {
		return isMassUpdate;
	}

	// Step definitions
	public abstract static class StepDTO<T extends StepDTO> {

		protected WorkflowDTO workflow;
		protected Map<String, Object> properties;
		protected String uniqueName;

		/** Signals that approvable should be exported and attached to email in the given step */
		protected Boolean emailAttachment = Boolean.FALSE;
		protected String emailAttachmentTemplateName = null;

		protected String alreadySatisfiedComment;

		public StepDTO(WorkflowDTO workflow, String uniqueName) {
			this.workflow = workflow;
			this.properties = new HashMap<>();
			this.uniqueName = uniqueName;
		}

		public Map<String, Object> getProperties() {
			return properties;
		}

		/**
		 * Gets the name of the step
		 * @return
		 */
		public String getUniqueName() {
			return uniqueName;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((uniqueName == null) ? 0 : uniqueName.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			StepDTO other = (StepDTO) obj;
			if (uniqueName == null) {
				if (other.uniqueName != null)
					return false;
			} else if (!uniqueName.equals(other.uniqueName))
				return false;
			return true;
		}

		/**
		 * Makes the step send the attachment along with the notification
		 * Same as <code>withEmailAttachment(true)</code>
		 */
		public T withEmailAttachment() {
			emailAttachment = Boolean.TRUE;
			return (T)this;
		}

		/**
		 * Makes the step send the attachment along with the notification
		 * @param emailAttachment Send attachment (true) or not (false)
		 */
		public T withEmailAttachment(Boolean emailAttachment) {
			this.emailAttachment = Boolean.TRUE.equals(emailAttachment);
			return (T)this;
		}

		/**
		 * Retrieves the attachment sending setting
		 * @return
		 */
		public Boolean getEmailAttachment() {
			return emailAttachment;
		}

		public Map<String, Object> toMap() {
			Map<String, Object> map = new HashMap<>();
			map.put("uniqueName", uniqueName);
			map.put("properties", properties);
			map.put("emailAttachment", emailAttachment);
			return map;
		}

		@SuppressWarnings("unchecked")
		public void fromMap(Map<String, Object> values) {
			if (values != null) {
				uniqueName = (String) values.get("uniqueName");
				properties = (Map<String, Object>) values.get("properties");
				emailAttachment = (Boolean) values.get("emailAttachment");
			}
		}

		protected List<String> readSingleValue(String name) {
			if (name == null) {
				return null;
			}

			return Collections.singletonList(name);
		}

		public T withEmailAttachmentTemplateName(String emailAttachmentTemplateName) {
			this.emailAttachmentTemplateName = emailAttachmentTemplateName;
			return (T)this;
		}

		public String getEmailAttachmentTemplateName() {
			return emailAttachmentTemplateName;
		}

		public String getAlreadySatisfiedComment() {
			return alreadySatisfiedComment;
		}

		/**
		 * Mark current step as already satisfied, workflowengine will skip its processing once marked as satisfied.
		 */
		public T withAlreadySatisfied() {
			return withAlreadySatisfied("");
		}

		/**
		 * Mark current step as already satisfied, workflowengine will skip its processing once marked as satisfied.
		 * @param comment If set to non-null value, the execution of the step will skipped and its value used as the step's comment
		 */
		public T withAlreadySatisfied(String comment) {
			this.alreadySatisfiedComment = comment;
			return (T)this;
		}
	}

	public static class ApprovalStepDTO extends StepDTO<ApprovalStepDTO> {

		private List<String> userLoginNames;
		private List<String> notificationOnlyToUsers;
		private List<String> userGroupNames;
		private Boolean excludeSubmitter;
		private List<String> reasons;
		private List<Object> approvalRequiredEmailAttachments;
		private List<String> denialReasons = new ArrayList<>();
		private boolean isDenialReasonOptional = true;
		private Integer minApprovalsNeeded = null;
		private List<Integer> minApprovalsForGroups = new ArrayList<>();
		private String postStepLogicName;
		private Boolean multiApprovalAllowed;
		private Boolean parallelStepApprovalAllowed;
		private List<String> mandatoryComments;

		public ApprovalStepDTO(WorkflowDTO workflow, String uniqueName) {
			super(workflow, uniqueName);
			approvalRequiredEmailAttachments = new ArrayList<>();
		}

		/**
		 * @deprecated Last invoked of {@code #withApprover(String)}, {@code #withUserGroupApprover(String)} wins
		 */
		@Deprecated
		public ApprovalStepDTO withApprover(String userLoginName) {
			this.userLoginNames = readSingleValue(userLoginName);
			this.userGroupNames = null;
			return this;
		}

		/**
		 * Last invoked of {@code #withApprovers(String...)}, {@code #withUserGroupApprovers(String...)} wins
		 */
		public ApprovalStepDTO withApprovers(String... userLoginNames) {
			this.userLoginNames = Arrays.asList(userLoginNames);
			this.userGroupNames = null;
			return this;
		}

		/**
		 * Subset of approvers that should receive the email notification
		 */
		public ApprovalStepDTO withNotificationOnlyToUsers(String... userLoginNames) {
			this.notificationOnlyToUsers = Arrays.asList(userLoginNames);
			return this;
		}

		/**
		 * @deprecated Last invoked of {@code #withApprover(String)}, {@code #withUserGroupApprover(LString)} wins
		 */
		@Deprecated
		public ApprovalStepDTO withUserGroupApprover(String userGroupUniqueName) {
			this.userGroupNames = readSingleValue(userGroupUniqueName);
			this.userLoginNames = null;
			return this;
		}

		/**
		 * Last invoked of {@code #withApprovers(String...)}, {@code #withUserGroupApprovers(String...)} wins
		 */
		public ApprovalStepDTO withUserGroupApprovers(String... userGroupUniqueNames) {
			this.userGroupNames = Arrays.asList(userGroupUniqueNames);
			this.userLoginNames = null;
			return this;
		}

		/**
		 * @deprecated Same as {@code #withUserGroupApprover(String)} but the submitter is excluded in case he is memeber of the group
		 * Last invoked of {@code #withApprover(String)}, {@code #withUserGroupApprover(String)} wins
		 */
		@Deprecated
		public ApprovalStepDTO withUserGroupApprover(String userGroupUniqueName, boolean excludeSubmitter) {
			this.userGroupNames = Collections.singletonList(userGroupUniqueName);
			this.userLoginNames = null;
			this.excludeSubmitter = excludeSubmitter;
			return this;
		}

		/**
		 * Set exclude submitter
		 */
		public ApprovalStepDTO withExcludeSubmitter(boolean excludeSubmitter) {
			this.excludeSubmitter = excludeSubmitter;
			return this;
		}

		/**
		 * Set an explanation reason for the approval step.
		 * @param reason The reason
		 * @return
		 */
		@Deprecated
		public ApprovalStepDTO setReason(String reason) {
			this.reasons = readSingleValue(reason);
			return this;
		}

		/**
		 * Sets explanation reasons for approvers of the approval step.
		 * @param reasons The reasons why each approver needs to approve. The reasons match approvers,
		 *                      e.g. if you have 3 approvers (A1, A2, A3), you can specify up to 3 reasons (R1, R2, R3)
		 *                      and then reason R1 applies to approver A1, reason R2 applies to approver A2, etc.
		 *                      If you want to put multiple reasons to a single approver, you need to join the reason
		 *                      messages together into one string and provide such single combined reason for the approver.
		 * @return The approval step itself.
		 */
		public ApprovalStepDTO withReasons(String... reasons) {
			this.reasons = Arrays.asList(reasons);
			return this;
		}

		public ApprovalStepDTO addProperty(String name, Object value) {
			this.properties.put(name, value);
			return this;
		}

		// exists due to end users convenience as it's possible to specify only one or more attachments. in this case via a filter or by similar means
		public ApprovalStepDTO withApprovalRequiredEmailAttachment(Object attachment) {
			return withApprovalRequiredEmailAttachments(attachment);
		}

		public ApprovalStepDTO withApprovalRequiredEmailAttachments(Object attachments) {
			approvalRequiredEmailAttachments.add(attachments);
			return this;
		}

		/**
		 * @deprecated Return the current approver in case of single approver set (or null if none set or if group approval)
		 * @return The user name
		 */
		@Deprecated
		public String getApprover() {
			return getSingleLoginName();
		}

		/**
		 * Return the current approvers (or null if none set or if group approval)
		 * @return The user names
		 */
		public List<String> getApprovers() {
			return userLoginNames;
		}

		/**
		 * Returns the subset of approvers that should recieve email notification.
		 * @return The user names
		 */
		public List<String> getNotificationOnlyToUsers() {
			return notificationOnlyToUsers;
		}

		/**
		 * Returns the reason text
		 * @return The reason
		 */
		@Deprecated
		public String getReason() {
			return getSingleReason();
		}

		/**
		 * Returns the reasons text
		 * @return The reasons
		 */
		public List<String> getReasons() {
			return reasons;
		}

		/**
		 * Returns status if the submitter is excluded in case of group approval
		 * @return true if excluded
		 */
		public Boolean getExcludeSubmitter() {
			return excludeSubmitter;
		}

		/**
		 * @deprecated Gets user group name in case of single group approval (or null otherwise)
		 * @return
		 */
		@Deprecated
		public String getUserGroupApprover() {
			return getSingleGroupName();
		}

		/**
		 * Gets user group names in case of group approval (or null otherwise)
		 * @return
		 */
		public List<String> getUserGroupApprovers() {
			return userGroupNames;
		}

		/**
		 * A minimum number of approvals needed to consider this step as approved. Default is 1
		 * @param minApprovalsNeeded
		 * @return
		 */
		public ApprovalStepDTO withMinApprovalsNeeded(int minApprovalsNeeded) {
			this.minApprovalsNeeded = minApprovalsNeeded;
			return this;
		}

		/**
		 * A minimum number of approvals needed for each group, to consider this step as approved. Default is 0 for each group.
		 * @param minApprovalsForGroups Resets minApprovalsNeeded
		 * @return
		 */
		public ApprovalStepDTO withMinApprovalsForGroups(Integer... minApprovalsForGroups) {
			this.minApprovalsForGroups = Arrays.asList(minApprovalsForGroups);
			return this;
		}

		/**
		 * Sets post step logic name that should be run after this step is executed (i.e. Approved, Denied or Withdrawn)
		 * @param logicUniqueName Workflow Post Step Logic unique name. If null, {@code WorkflowDTO#withDefaultPostApprovalStepLogic(String)} will be used,
		 *                        if empty string, no logic will be used even if there is {@code WorkflowDTO#withDefaultPostApprovalStepLogic(String)} set
		 */
		public ApprovalStepDTO withPostStepLogic(String logicUniqueName) {
			this.postStepLogicName = logicUniqueName;
			return this;
		}

		public String getPostStepLogicName() {
			return postStepLogicName;
		}

		public List<Object> getApprovalRequiredEmailAttachments() {
			return approvalRequiredEmailAttachments;
		}

		private String getSingleLoginName() {
			return userLoginNames != null && userLoginNames.size() == 1 ? userLoginNames.get(0) : null;
		}

		private String getSingleGroupName() {
			return userGroupNames != null && userGroupNames.size() == 1 ? userGroupNames.get(0) : null;
		}

		private String getSingleReason() {
			return reasons != null && reasons.size() == 1 ? reasons.get(0) : null;
		}

		@Override
		public Map<String, Object> toMap() {
			Map<String, Object> map = super.toMap();
			map.put("userLoginName", getSingleLoginName());
			map.put("userGroupName", getSingleGroupName());
			map.put("userLoginNames", userLoginNames);
			map.put("userGroupNames", userGroupNames);
			map.put("excludeSubmitter", excludeSubmitter);
			map.put("reason", getSingleReason());
			map.put("reasons", reasons);
			map.put("approvalRequiredEmailAttachments", approvalRequiredEmailAttachments);
			map.put("minApprovalsNeeded", minApprovalsNeeded == null ? 1 : minApprovalsNeeded);
			map.put("minApprovalsForGroups", minApprovalsForGroups);
			map.put("postStepLogicName", postStepLogicName);
			map.put("multiApprovalAllowed", multiApprovalAllowed);
			map.put("parallelStepApprovalAllowed", parallelStepApprovalAllowed);
			map.put("mandatoryComments", mandatoryComments);
			return map;
		}

		@SuppressWarnings({ "unchecked", "rawtypes" })
		@Override
		public void fromMap(Map<String, Object> values) {
			if (values != null) {
				super.fromMap(values);

				userLoginNames = readSingleValue((String) values.get("userLoginName"));
				userGroupNames = readSingleValue((String) values.get("userGroupName"));
				userLoginNames = Optional.ofNullable((List<String>) values.get("userLoginNames")).orElse(userLoginNames);
				userGroupNames = Optional.ofNullable((List<String>) values.get("userGroupNames")).orElse(userGroupNames);
				excludeSubmitter = (Boolean) values.get("excludeSubmitter");
				reasons = readSingleValue((String) values.get("reason"));
				reasons = Optional.ofNullable((List<String>) values.get("reasons")).orElse(reasons);
				approvalRequiredEmailAttachments = (List) values.get("approvalRequiredEmailAttachments");
				minApprovalsNeeded = Optional.ofNullable((Integer) values.get("minApprovalsNeeded")).orElse(1);
				minApprovalsForGroups = Optional.ofNullable((List<Integer>) values.get("minApprovalsForGroups")).orElse(Collections.emptyList());
				postStepLogicName = (String) values.get("postStepLogicName");
				multiApprovalAllowed = (Boolean) values.get("multiApprovalAllowed");
				parallelStepApprovalAllowed = (Boolean) values.get("parallelStepApprovalAllowed");
				mandatoryComments = Optional.ofNullable((List<String>) values.get("mandatoryComments")).orElse(Collections.emptyList());
			}
		}

		public ApprovalStepDTO withDenialReasons(List<String> denialReasons, boolean isDenialReasonOptional) {
			this.denialReasons.addAll(denialReasons);
			this.isDenialReasonOptional = isDenialReasonOptional;
			return this;
		}

		public List<String> getDenialReasons() {
			return denialReasons;
		}

		public boolean isDenialReasonOptional() {
			return isDenialReasonOptional;
		}

		public Integer getMinApprovalsNeeded() {
			return minApprovalsNeeded;
		}

		public List<Integer> getMinApprovalsForGroups() {
			return minApprovalsForGroups;
		}

		public Boolean isMultiApprovalAllowed() {
			return multiApprovalAllowed;
		}

		public Boolean isParallelStepApprovalAllowed() {
			return parallelStepApprovalAllowed;
		}

		public List<String> getMandatoryComments() { return mandatoryComments; }

		@Deprecated
		public ApprovalStepDTO withMultipleApprovalByGroupsAllowed(boolean multipleApprovalByGroupsAllowed) {
			withSingleUserMultiApprovalAllowed(multipleApprovalByGroupsAllowed);
			return this;
		}

		/**
		 * @param multiApprovalAllowed When true, then an approver can approve not only once, but that many times
		 * as he has number of groups that are allowed to approve.
		 */
		public ApprovalStepDTO withSingleUserMultiApprovalAllowed(boolean multiApprovalAllowed) {
			this.multiApprovalAllowed = multiApprovalAllowed;
			return this;
		}

		/**
		 * @param parallelStepApprovalAllowed When true, multiApprovalAllowed enabled, the approver will by single
		 * approve trigger approve all all parallel steps that he is entitled to.
		 */
		public ApprovalStepDTO withParallelStepApprovalAllowed(boolean parallelStepApprovalAllowed) {
			this.parallelStepApprovalAllowed = parallelStepApprovalAllowed;
			return this;
		}

		/**
		 * Makes comment mandatory for certain actions. <p>
		 * Example:
		 * <pre>{@code workflow.addApprovalStep("ApproverWithDefault")
		 *	.withUserGroupApprovers("Group1", "Group2", "Group3")
		 *	.withMandatoryComments("DENY", "APPROVE")
		 *	.withMinApprovalsForGroups(1, 1, 2)
		 * }</pre>
		 * </p>
		 * @param actions for which comment is mandatory. This method accepts only values "APPROVE" and "DENY". Other values will be ignored.
		 * @return
		 */
		public ApprovalStepDTO withMandatoryComments(String... actions) {
			this.mandatoryComments = Arrays.asList(actions).stream().map(String::toUpperCase).collect(Collectors.toList());
			this.mandatoryComments.retainAll(Arrays.asList("APPROVE", "DENY"));
			return this;
		}
	}

	public static class WatcherStepDTO extends StepDTO<WatcherStepDTO> {

		private List<String> userLoginNames;
		private List<String> userGroupNames;
		private String[] emails;
		private List<String> reasons;

		public WatcherStepDTO(WorkflowDTO workflow, String userLoginName) {
			super(workflow, userLoginName);
		}

		/**
		 * @deprecated Specify a user as recipient
		 * Last invoked of {@code #withUserWatcher(String)}, {@code #withUserGroupWatcher(String)},
		 * {@code #withEmailListWatcher(String...)} wins
		 */
		@Deprecated
		public WatcherStepDTO withUserWatcher(String userLoginName) {
			this.userLoginNames = readSingleValue(userLoginName);
			this.userGroupNames = null;
			this.emails = null;
			return this;
		}

		/**
		 * Specify users as recipients
		 * Last invoked of {@code #withUserWatchers(String...)}, {@code #withUserGroupWatchers(String...)},
		 * {@code #withEmailListWatcher(String...)} wins
		 */
		public WatcherStepDTO withUserWatchers(String... userLoginNames) {
			this.userLoginNames = Arrays.asList(userLoginNames);
			this.userGroupNames = null;
			this.emails = null;
			return this;
		}

		/**
		 * @deprecated Specify user group as recipient
		 * Last invoked of {@code #withUserWatcher(String)}, {@code #withUserGroupWatcher(String)},
		 * {@code #withEmailListWatcher(String...)} wins
		 */
		@Deprecated
		public WatcherStepDTO withUserGroupWatcher(String userGroupUniqueName) {
			this.userGroupNames = readSingleValue(userGroupUniqueName);
			this.userLoginNames = null;
			this.emails = null;
			return this;
		}

		/**
		 * Specify user groups as recipient
		 * Last invoked of {@code #withUserWatchers(String...)}, {@code #withUserGroupWatchers(String...)},
		 * {@code #withEmailListWatchers(String...)} wins
		 */
		public WatcherStepDTO withUserGroupWatchers(String... userGroupUniqueNames) {
			this.userGroupNames = Arrays.asList(userGroupUniqueNames);
			this.userLoginNames = null;
			this.emails = null;
			return this;
		}

		/**
		 * Instead of specifying a user name or a user group, arbitrary email addresses can be used here
		 * Last invoked of {@code #withUserWatchers(String...)}, {@code #withUserGroupWatchers(String...)},
		 * {@code #withEmailListWatcher(String...)} wins
		 */
		public WatcherStepDTO withEmailListWatcher(String... emails) {
			this.emails = emails;
			this.userLoginNames = null;
			this.userGroupNames = null;
			return this;
		}

		public WatcherStepDTO addProperty(String name, Object value) {
			this.properties.put(name, value);
			return this;
		}

		public WatcherStepDTO setReason(String reason) {
			this.reasons = readSingleValue(reason);
			return this;
		}

		public WatcherStepDTO setReason(List<String> reasons) {
			this.reasons = reasons;
			return this;
		}

		 /**
		 * Sets explanations for the watcher step.
		 * @param reasons The reasons why each watcher is notified.
		 * The reasons match watchers. E.g. if you have 3 watchers (w1, w2, w3),
		 * you can specify up to 3 reasons (r1, r2, r3) and then r1 corresponds
		 * to w1, r2 to w2 etc. If you want to put multiple reasons to a single watcher,
		 * you need to join the strings and return a single combined reason in the code.
		 * @return
		 */
		public WatcherStepDTO withReasons(String... reasons) {
			this.reasons = Arrays.asList(reasons);
			return this;
		}

		/**
		 * @deprecated Return the current watcher (or null if none set or if group watcher)
		 * @return The user name
		 */
		public String getUserWatcher() {
			return getSingleLoginName();
		}

		/**
		 * Return the current watchers (or null if none set or if group watchers)
		 * @return The user names
		 */
		public List<String> getUserWatchers() {
			return userLoginNames;
		}

		/**
		 * Returns the reason text
		 * @return The reason
		 */
		@Deprecated
		public String getReason() {
			return getSingleReason();
		}

		/**
		 * Returns the reason text
		 * @return The reason
		 */
		public List<String> getReasons() {
			return reasons;
		}

		/**
		 * @deprecated Gets user group name in case of group watcher (or null otherwise)
		 * @return
		 */
		public String getUserGroupWatcher() {
			return getSingleGroupName();
		}

		/**
		 * Gets user group names in case of group watchers (or null otherwise)
		 * @return
		 */
		public List<String> getUserGroupWatchers() {
			return userGroupNames;
		}

		/**
		 * Gets the list of email addresses if specified as recipient for this node
		 * @return
		 */
		public String[] getEmailListWatcher() {
			return emails;
		}

		private String getSingleLoginName() {
			return userLoginNames != null && userLoginNames.size() == 1 ? userLoginNames.get(0) : null;
		}

		private String getSingleGroupName() {
			return userGroupNames != null && userGroupNames.size() == 1 ? userGroupNames.get(0) : null;
		}

		private String getSingleReason() {
			return reasons != null && reasons.size() == 1 ? reasons.get(0) : null;
		}

		@Override
		public Map<String, Object> toMap() {
			Map<String, Object> map = super.toMap();
			map.put("userLoginName", getSingleLoginName());
			map.put("userGroupName", getSingleGroupName());
			map.put("userLoginNames", userLoginNames);
			map.put("userGroupNames", userGroupNames);
			map.put("emails", emails);
			map.put("reason", getSingleReason());
			map.put("reasons", reasons);
			return map;
		}

		@SuppressWarnings("unchecked")
		@Override
		public void fromMap(Map<String, Object> values) {
			if (values != null) {
				super.fromMap(values);

				userLoginNames = readSingleValue((String) values.get("userLoginName"));
				userGroupNames = readSingleValue((String) values.get("userGroupName"));
				userLoginNames = Optional.ofNullable((List<String>) values.get("userLoginNames")).orElse(userLoginNames);
				userGroupNames = Optional.ofNullable((List<String>) values.get("userGroupNames")).orElse(userGroupNames);

				Object emailsRaw = values.get("emails");
				if (emailsRaw != null) {
					if (emailsRaw instanceof String[]) {
						emails = (String[]) emailsRaw;
					} else if (emailsRaw instanceof List) {
						emails = ((List<String>) emailsRaw).toArray(new String[0]);
					} else {
						throw new IllegalArgumentException("emails: " + emails);
					}
				}
				reasons = readSingleValue((String) values.get("reason"));
				reasons = Optional.ofNullable((List<String>) values.get("reasons")).orElse(reasons);
			}
		}
	}
}
