package net.pricefx.formulaengine;

import com.googlecode.genericdao.search.Filter;
import net.pricefx.common.api.FieldFormatType;
import net.pricefx.common.util.CalendarUtil.TimePeriod;
import net.pricefx.domain.FilterNode;
import net.pricefx.formulaengine.scripting.Matrix2D;
import net.pricefx.formulaengine.DatamartContext;

import java.io.IOException;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * The ModelFormulaContext is exposed to model logics through the <code>model</code> binding.  It provides access to a PO specific API,
 * not available in any of the other Pricefx modules.  API calls that add, update or delete objects are only possible in a model calculation step.
 * <p>
 * When called in the 'Definition', 'Analysis' or 'Evaluation' context, no action will be taken, and if any value is to be returned, this value will be <code>null</code>.
 * <br>A PO model has at least one stage: 'Definition'.  This is the context of the PO logic called when saving or updating a model.
 * <p>
 * It should be noted that the UI updates a model before triggering a calculation (though this is client specific behaviour, not required or enforced
 * by the Pricefx backend).  A model is therefore typically updated multiple times in its life cycle,
 * as when first created, it tends to be (re-)calculated many times thereafter.
 * <p>
 * While not mandatory, a PO model is not likely to be useful, or even needed, when no Calculation steps are defined.
 * Costly, i.e. computationally expensive, time consuming jobs are normally required to be run to prepare a model for future use.
 * In machine learning terminology, one might refer to this as the 'training' phase.
 * Once trained, one might talk about using the model for prediction, or classification.  In Pricefx this is achieved by evaluating
 * the model.  While, again, optional, a useful model tends to define an 'Evaluation' logic that performs this function.
 * <p>
 * The ModelType is what pieces this together.  Models are created from a chosen ModelType.
 * <br>The <code>ModelType</code> defines these key characteristics:
 * </p>
 * <p><ul>
 *     <li><b>logic:</b>    UniqueName of the main logic;  this is the logic that as minimum implements the 'Definition' behaviour.
 *                          It can also, selectively, define the behaviour in other stages by virtue of setting the 'Calculation Context' on the logic element.
 *                          (see further down for a list of valid calculation contexts).
 *     <li><b>libLogic:</b> Optional.  UniqueName of a logic that serves as a library specific to this ModelType.  Its elements can be accessed through
 *                          the <code>lib</code> binding.
 *     <li><b>analysisLogic:</b>    Optional.  UniqueName of the logic that is evaluated when triggering an analysis or inspection from the UI.  What this looks like is
 *                          specific to the UI.  What is important is that it is a triggered by a <code>datamart.modeleval/analysis</code> request to
 *                          the JSON API.
 *     <li><b>evalLogic:</b>  Optional.  UniqueName of the logic that is evaluated when using ('evaluating') a prepared model from other modules, for example from
 *                          Pricing logic in QuoteConfigurator or PriceBuilder.  See {@link ModelEvaluator} for details.
 *     <li><b>recordLogic:</b> Optional.  UniqueName
 *     <li><b>generatesData:</b>    True or false.  Indicates that the module will use its PA backing table to store data, result etc.
 *                          A client can use this flag to decide on whether to show a data table for the model or not.
 *     <li><b>generatesTables:</b>  True or false.  Indicates that the module will create additional DMTables.
 *                          A client can use this flag to decide on whether to show a section where these tables can be accessed.
 *     <li><b>generatesRecords:</b> True or false.  Indicates that the module will generate ModelRecords (the technical name) aka 'Policy Records' (the UI name).
 *                          A client can use this flag to decide on whether to show a section where these records can be accessed.
 *     <li><b>generatesTrees:</b>   True or false.  Indicates that the module will generate one or more trees (see {@link #addTree(String, Collection)}).
 *                          A client can use this flag to decide on whether to show a section where these trees can be accessed.
 *     <li><b>calcSteps:</b>    Optional.  Defines a List<CalcStep>, of which each can be triggered independently and in any other, though the expectation
 *                          is that they would, at least initially, be executed in sequence.  Some steps, depending entirely on the PO logic implementation,
 *                          might be skipped, and some steps might be rerun after the initial calculation.  There are no enforced rules; the ModelType
 *                          documentation should clarify the allowed and desired behaviour.
 * </ul></p>
 * A <code>CalcStep</code> is defined as:
 * <p><ul>
 *     <b><li>name:</b>     The name of the calculation step, to be unique across all steps.  This is the value returned by a {@link #getCalculationStep()} call.
 *     <b><li>label:</b>    An optional label or description that can be used by the UI.
 *     <b><li>category:</b> An optional category, that allows a UI to group calculation steps together.  In the Unity PO UI, for the Segmentation model type,
 *                          calculation steps are divided in 'Analysis' steps (not to be confused with the analysis logic behaviour) and 'Calculation' steps.
 *                          In Segmentation, the 'Analysis' steps are "DataPrep', 'DataProfile' and 'PriceDrivers'.  The 'Calculation' steps are
 *                          'Segmentation', 'Scoring' and 'Pricing'.
 *     <b><li>logic:</b>    The optional UniqueName of the logic to be executed for this step.  When not set, the calculation step's logic is assumed
 *                          to be implemented by the (mandatory) main logic specified in the <code>ModelType.logic</code> attribute.
 *     <b><li>distributed:</b>  True or false.  Indicates whether this step can be executed in distributed mode.
 *                          For a non-distributed step, no additional logic level configuration is required.  PO will simply execute this step
 *                          using the associated logic.  For a distributed step, however, there is a pre-condition in that some previous step in the
 *                          model calculation is assumed to have created so called <code>CalculationItems</code>, representing units of work for this step.
 *                          Each such unit of work can then be executed by one of the workers in a distributed execution pool.  There is in fact a threshold,
 *                          below which the distributed calculation will not kick in (100 by default, set by the <code>formulaEngine.script.optimizer.slaveCalcItemThreshold</code>
 *                          instance parameter), in order to avoid creating overhead for low number of work units.
 *                          See {@link #addOrUpdateCalcItem(Object, Object, Object)} and its variations for more details.
 * </ul></p>
 *
 * The list of Calculation Contexts is
 * <p><ul>
 *     <b><li>definition:</b>               Evaluated when a model is saved
 *     <b><li>calculation:</b>              Evaluated in a calculation step
 *     <b><li>calculation-init:</b>         Evaluated when calculating a distributed step, once, before the individual CalculationItems are processed
 *     <b><li>calculation-summary:</b>      Evaluated when calculating a distributed step, once, after all the individual CalculationItems have been processed
 *     <b><li>analysis:</b>                 Evaluated in in the Analysis (Classic) and Results (Unity) sections of the UI
 *     <b><li>evaluation:</b>               Evaluated when {@link ModelEvaluator#evaluate(Map)} is invoked
 *     <b><li>record:</b>                   Evaluated after a record using {@link #addRecord(String, TimePeriod, Map, Map, boolean...)} when the <code>calculate == true</code>
 *   </ul>
 * </p></pre>
 */
public interface ModelFormulaContext extends DistFormulaContext {

    /**
     * @return  The unique name of the model
     */
    String getUniqueName();

    /**
     * @return  The label name of the model
     */
    String getLabel();

    /**
     * Returns the source name of this model, which is often needed as a parameter in queries, charts and such on the model in the analysis phase;
     * and avoids retreiving the whole model with get().
     * <br>Typically used when building a query against the model's backing table, and when defining charts with the model as a source.
     * <p>
     * In the 'Segmentation' logic for example, the 'DataPrep' step filters and loads the transaction data in the model's table, and subsequent
     * steps then query the model's table for better performance.
     * <p>
     * Note: The source name for a PO model has the form 'DMM.'<uniqueName>.
     * </p>
     * @return The source name of the model
     */
    String getSourceName();

    /**
     * A typical (set of)  PO logic(s) covers various stages:  definition, calculation, analysis, record and evaluation.
     * <br>Not all are always needed - the ModelType defines which one are.  Still, a persisted data object accessible to all stages is usually required.
     * <br>In the 'Segmentation' model type, for example, the selected transactions source and field mapping (product, customer etc. fields) are stored
     * by the Definition logic in the built-in <code>model.def</code> map.
     * <p>
     * Note:  As the map is persisted in the model, it is important not to put any non-serializable objects in it.
     *        When in doubt, stick to Collections of (Collections of...) primitives, String, Number, Date etc.
     * </p>
     * Sample code:
     * <p><pre><code>
     * model.def.policyStartDate = api.dateUserEntry("Start Date")
     * model.def.policyEndDate = api.dateUserEntry("End Date")
     * </code></pre></p>
     * @return  The model.def map
     */
    Map<String, Object> getDef();

    /**
     * When a clients retrieves the required input parameters for a given stage (definition, calculation step, analysis...), there is the option
     * to set an initial value for a parameter.  This is done with code like this snippet:
     * <p><pre><code>
     *  def PARAM = "Floor Percentile"
     *  model.def.floorPercentile = api.integerUserEntry(PARAM)
     *  model.defaults[PARAM] = 50
     * </code></pre></p>
     * In a new model, the UI will render this input having the initial 50 value, but then the user can override it and save.
     * @return  The map of default parameters
     */
    Map<String, Object> getDefaults();

    /**
     * Exposes all the input values passed in with a ModelEval request (user entries), or all of the inputs persisted
	 * in the model, when in a calculation.  Note that in the ModelEval scenario, the inputs are not necessarily stored
	 * in the model yet.  In this use case, the logic leverages knowledge of already provided input values (provided
	 * by the client/UI) to determine additionally required inputs.
     * @return	A map of input parameter (name,value) pairs.
     */
    Map<String, Object> getInput();

        /**
		 * Each model is backed by a table in the PA DB.  This way, a model can be used as a source to a query or chart.
		 * <br>But unlike for example a DataSource, the user or configuration engineer is not expected to define its fields in the UI.
		 * <br>Instead, the model dynamically - in the definition logic - defines its fields using this method.  The fields to be added
		 * normally depend on user inputs, such as the selected revenue, volume and margin fields in the Segmentation logic example.
		 * <p>Sample code:</p>
		 * <p><pre><code>
		 *  def ctx = api.getDatamartContext()
		 *  def sourceName = ctx.sourceSelectionEntry("Transaction Source", "DM", "DMDS")    // allow selection from DMs and DSs
		 *  def source = ctx.getFieldCollection(sourceName)
		 *  if (source) {
		 *      def fields = source.fc.fields.collectEntries{ [it.name, it] }
		 *      def fieldName = ctx.fieldSelectionEntry("Product Field", source, "DIM")
		 *      if (fieldName) {
		 *          def field = fields[fieldName]
		 *          model.addField(field)      // leverage the field definition from the source
		 *          model.addField( [name:field.name, label:field.label, type:field.type, dimension:true, key:false] )    // or define it explicitly
		 *      }
		 * </code></pre></p>
		 * @param fieldDef  The field definition as a map of (property, values) pairs
		 * @return  The model binding. to facilitate a builder pattern
		 */
    ModelFormulaContext addField(Object fieldDef);

    /**
     * Retrieves the field definitions previously added using {@link #addField(Object)}.
     * <br>Note that as DMField is a non-whitelisted type, the Groovy sandbox will convert the DMField objects to maps, resulting in the
     * same field definition representation as used in {@link #addField(Object)}.
     * @return Collection of field definitions previously added using {@link #addField(Object)}
     */
    Collection<Object> getFields();

    DatamartContext.Query getSourceQuery();
	void setSourceQuery(DatamartContext.Query query);


	/**
	 * A ModelType definition, optionally (but usually) defines a number of calculation steps.  When the user, or a 3rd party via the JSON API,
	 * or a schedule ModelCalculation DataLoad, request a calculation, the step to execute is identified by the DataLoad's <code>calculationContext</code>
	 * attribute.  This allows PO to identify the step, and hence the logic to be executed.  When there is no dedicated logic defined for each calculation step,
	 * this method exposed the <b>name</b> of the calculation step being executed to the logic, so it can selectively evaluate the relevant elements.
	 * @return  A map representation of the <code>CalcStep</code> being executed.  Usually <code>CalcStep.name</code> is sufficient to drive any conditional
	 * Groovy code execution.
	 */
    Object getCalculationStep();

    /**
     * Adds a tree to this model, which is effectively a FilterNode object that acts as the root of the tree, i.e. one that does not point
     * to a parent FilterNode.  The <code>levelNames</code> are persisted in the FilterNode.output attribute, later used to construct a more readable
     * map representation of the node's path (see {@link #getFilterNodePath()} and {@link #getFilterNodePath(Long)}.
     * <br>If a tree with the given name already exists, the id of the root node is returned, and no further action is taken.
     * @param treeName  Name to be set on the root node (FilterNode.name)
     * @param levelNames Names of the first, second etc. level of the tree
     * @return The id of the root node of the tree
     */
    Long addTree(String treeName, Collection<String> levelNames);

    /**
     * Adds a tree to this model, consisting of nodes constructed from a data set, usually a PA query result.
     * <br>The root node is the same as created by {@link #addTree(String, Collection)}.
     * <br>The immediate descendents of the root node are created from the values found in the first level column (as dictated by the <code>levelCols</code> argument).
     * <br>The second level descendants are created from the values in the second level column, if any.  And so on.
     * <p>
     * Example: creating a tree with name 'Segments' from data set
     * <p><pre>
     * Region,	Channel
     * EMEA,	Retail
     * APAC,	Retail
     * APAC,	Industry
     * </pre></p>
     * results in the following tree, when the level columns are chosen to be 'Region' and 'Channel', in that order.
     * <p><pre>
     * Segments
     *      |
     * 	 + EMEA
     * 	 |    |
     * 	 |    + Retail
     * 	 |
     * 	 + APAC
     * 	      |
     * 		  + Retail
     * 		  |
     * 		  + Industry
     * </pre></p>
     * <br>If a tree with the given name already exists, it is deleted before the new one is created.
     * @param treeName  Name of the tree to be created, i.e. FilterNode.name of the root node
     * @param data      Data from which to construct the tree
     * @param levelCols   Labels of the columns in <code>data</code> to be used to 1) name the tree levels and 2) identify the column to populate each level from
     * @return  All of the added tree nodes' ids
     */
    List<Long> addTree(String treeName, Matrix2D data, Collection<String> levelCols);

    /**
     * Creates or replaces a tree as in {@link #addTree(String, Matrix2D, Collection)}, with the additional specification of an output
     * or output format, and/or labels.  Whereas in the previous method tree nodes are created with FilterNode.output == null and Filter.label -- null,
     * this method sets:
     * <p><ul>
     *     <li>output field to the value found in the column identified by outputColName (== column label).  If outputColName is <code>null</code>
     *     and outputColFormat is not, then the output value is constructed as <code>String.format(outputColFormat, n);</code>, with <code>n</code>
     *     being the (n-1)th node created (i.e. <code>n=0</code> for the root node).
     *     <li>label field to <code>"[val1, val2, ...]"</code> where <code>val1, val2, ...</code> are the values found in the columns identified
     *     by <code>labelCols</code> respectively.
     * </ul></p>
     *
     * @param treeName  Name of the tree to be created, i.e. FilterNode.name of the root node
     * @param data  Data from which to construct the tree
     * @param levelCols Labels of the columns in <code>data</code> to be used to 1) name the tree levels and 2) identify the column to populate each level from
     * @param labelCols Labels of the columns in <code>data</code> from which to construct the label (sets FilterNode.label)
     * @param outputColName Label of the column in <code>data</code> from which to pick the output value (sets FilterNode.output)
     * @param outputColFormat   The format from which to construct the output value when <code>outputColName == null</code>
     * @return  All of the added tree nodes' ids
     */
    List<Long> addTree(String treeName, Matrix2D data, Collection<String> levelCols, Collection<String> labelCols, String outputColName, String outputColFormat);

    /**
     * Deletes the tree with the given name, if it exists.  No action is taken and no error returned if it doesn't.
     * Note that deleting a tree means deleting the root node as well as all of its direct and indirect descendents.
     * @param treeName  Name of the tree, i.e. name of root FilterNode, to delete
     */
    void deleteTree(String treeName);

    /**
     * Adds a descendent to an existing tree node, identified by its <code>parentId</code>.
     * <br>If the parent node does not exist, no action is taken and no error is returned.
     * @param parentId  Id of the parent node
     * @param name  To be FilterNode.name  (of the added node)
     * @param label To be FilterNode.label
     * @param output    To be FilterNode.output
     * @param filter    To be FilterNode.filter
     * @return  The id of the added tree node, or <code>null</code> if the parent node doesn't exist
     */
    Long addTreeNode(Long parentId, String name, String label, String output, Filter filter);

    /**
     * Finds the tree node, if it exists, in the given tree with the given <code>path</code>.
     * <br>The path is the list of values for the first, second etc. level of the tree (see e.g. {@link #addTree(String, Collection)}).
     * <br>Note: to find a leaf node, values for all levels are to be provided.  Fewer values will result in an intermediate level node
     * to be returned, if such a node exists of course.
     * @param treeName  Name of the tree to be created, i.e. FilterNode.name of the root node
     * @param path  List of values for the various levels in the tree.
     * @return  The id of the FilterNode with the given path (from the root node).  See also {@link #getFilterNodePath(Long)}
     */
    Long findTreeNode(String treeName, List<String> path);

    /**
     * Given the id of a FilterNode, in some tree previously added to the model, this method returns the logical path from the root to that node.
     * <p>
     * Example:  if the identified node in the below tree is 'Retail' in EMEA, then the filter node path is [Region:"EMEA", Channel:"Retail"],
     * with 'Region' and 'Channel' assumed to be the level names, in that order, defined when first creating the tree (see {@link #addTree(String, Collection)}).
     * <p><pre>
     * Segments
     *      |
     * 	 + EMEA
     * 	 |    |
     * 	 |    + Retail
     * 	 |
     * 	 + APAC
     * 	      |
     * 		  + Retail
     * 		  |
     * 		  + Industry
     * </pre></p>
     * @param nodeId    The id of a FilterNode that exists in a tree previously added to the model
     * @return  The path to the given FilterNode as a map of (level,value) pairs
     */
    Map<String, Object> getFilterNodePath(Long nodeId);

    /**
     * In the context of a call to the <code>datamart.modeleval</code> end point (JSON API), the client request can specify the 'filterNodeId' parameter.
     * <br>The value of this parameter is to be the id of a FilterNode object, previously created by the model.  In the 'Segmentation' model type, for example,
     * the 'Segmentation' calculation step creates a 'Segments' tree, in which each node, including the root node, is a FilterNode object.
     * <br>In the 'Results' section in the Unity UI, when the user selects a node in the Segments tree, a subsequent <code>modeleval</code> call retrieves the details
     * of the selected segment.
     * @return  A map representation of the selected FilterNode
     */
    FilterNode getFilterNode();

    /**
     * Like {@link #getFilterNodePath(Long)}, but for the preset FilterNode revealed by {@link #getFilterNode()} only.
     * @return  The path to the selected FilterNode as a map of (level,value) pairs
     */
    Map<String, Object> getFilterNodePath();

    /**
     * Adds a new record with the given name if none exist yet with that name in the DRAFT status (<code>workflowStatus == WorkflowStatus.DRAFT</code>).
     * Otherwise, that existing record is updated instead.
     * Note that there can be multiple records with the same name, as long as there is only one in the DRAFT state.  Usually these records will then also
     * have varying validity periods (for ex. one valid and approved for Q1, with a new one, yet to be approved for Q2).
     * @param name  Name of the record to be added/updated.
     * @param period    The validity period for the record, i.e. defining its <code>startDate</code> and <code>endDate</code> (both inclusive)
     * @param attributeValues   Map of attribute (label, value) pairs.  Note that this is a map from the field's label (not name)
     * @param fieldFormatTypesByColLabel    Map of attribute (label, {@link FieldFormatType}), defining the type of each record attribute to be set.
     *                                      Entries in the <code>attributeValues</code> argument for which no <code>FieldFormatType</code> is defined, are ignored, i.e. are not set.
     * @param calculate Optional flag, indicating that once the record is added/updated, a second, embedded, evaluation takes place, in which the record
     *                  being calculated is identified by the result of {@link #getRecord()}.  See also the <code>MethodType.recordLogic</code> property.
     * @return  A map representation of the added/updated ModelRecord
     */
    Object addRecord(String name, TimePeriod period, Map<String, ?> attributeValues, Map<String, FieldFormatType> fieldFormatTypesByColLabel, boolean... calculate);

    /**
     * Gets the ModelRecord ('PolicyRecord' in the UI) set as context in the 'record' logic evaluation (the logic set in the ModelType.recordLogic
     * attribute).  The evaluation is done when a record is added using {@link #addRecord(String, TimePeriod, Map, Map, boolean...)}
     * when the calculate argument is true.
     * @return  A map representation of the ModelRecord
     */
    Map<String, Object> getRecord();

    /**
     * Deletes all ModelRecords associated with the model, which are not yet submitted into the workflow.
     * @return  The number of records created.
     */
    int deleteRecords();

    /**
     * Finds the ModelRecord, if it exists, with the given name, valid on the given day (record.startDate &lt;=  validOnDate &lt;= record.endDate),
     * and with the given workflow status, if not <code>null</code>.  If wfStatus is <code>null</code>, there is no filter condition on record.workStatus.
     * <p>
     * Valid workflow statuses are:
     * <ul>
     *     <li>DRAFT
     *     <li>SUBMITTED
     *     <li>DENIED
     *     <li>APPROVED
     *     <li>NO_APPROVAL_REQUIRED
     *     <li>WITHDRAWN
     * </ul></p>
     * @param name  The name of the record give when first creating it, using {@link #addRecord(String, TimePeriod, Map, Map, boolean...)}}
     * @param validOnDate   The date on which the record is to be valid
     * @param wfStatus  The workflow status the records needs to be in, ignored if <code>null</code>
     * @return  A map representation of the record
     */
    Map<String, Object> findRecord(String name, Date validOnDate, String wfStatus);

    /**
     * Finds ModelRecords associated to this model using only Filter criteria.
     * @param filter    The criteria the records needs toad here to
     * @param sortBy    Optional sort fields
     * @return  A list of map representations of the records, empty of none matched the criteria.
     */
    List<Map<String, Object>> findRecords(Filter filter, String... sortBy);

    /**
     * Creates a new or updates an existing calculation item (entry in the Model CalcItems PP) for a given calculation step.
     * For example, a  'Segmentation' step may create a CalcItem for scoring 'Segment 001', injected into the logic context when the calculation
     * of the 'Scoring' step for that segment is initiated.
     * @param calcStep  The name of the calculation step in which the added item is to be calculated;  this goes into <code>Key1</code>
     * @param id    <Code>Key2</Code> of the calculation item
     * @param value A JSON-serializable value to be stored in the <code>Value</code> attribute of the calculation item
     * @return
     */
    Object addOrUpdateCalcItem(Object calcStep, Object id, Object value);

    /**
     * Gets all the entries of the LookupTable (PP) named 'Model Params &lt;model.uniqueName&gt;'.
     * All the user entries for the model are automatically stored in that PP.
     * In addition, the logic can add parameters programmatically (cf. {@link #setParam(String, Object, boolean...)}).
     * @return  A map of param (name, value) pairs
     */
    Map<String, Object> getParams();

    /**
     * Adds a new parameter, or overwrites an existing parameter value, being an entry in the 'Model Params &lt;model.uniqueName&gt;' PP.
     * @param name  Name of the parameter
     * @param value The parameter value, which needs to be JSON-serializable
     * @param ifMissing If true, an existing parameter's value will not be overwritten. This makes it easier to add default parameters,
     *                  not overwriting the value that was set by an end user.
     * @return  The added PP entry, as a map.  Mostly useful to get the <code>id</code> of the added/updated entry.
     */
    Object setParam(String name, Object value, boolean... ifMissing);

    /**
     * @return The names of all the files attached to the model.  This can for example be used to populate a dropdown user entry.
     * @throws IOException
     */
    List<String> getAttachments() throws IOException;

    /**
     * Creates a table, represented by a DMTable object in the PA DB, modelled on and populated from the given source.
     * <br>The uniqueName of the table is a combination of the model's typedId and the passed in tableName.
     * The table's label is set to the given tableName or from the source.
     * <br>If a DMTable with this tableName (i.e. with equal derived uniqueNames) already exists, that table is updated
     * and truncated rather than dropped and recreated.
     * If this is not desirable, then the existing table can be dropped beforehand.
     * <br>Note that when directly using a {@link net.pricefx.formulaengine.DatamartContext.SqlQuery}, the types of the selected fields
     * must be made explicit via casting for the resulting table to have the proper types.
     * <br>If a query is passed as a source, the projections become the fields of the new/updated table, and the result
     * is used to populate the table.
     * <p>
     * Note: Can only be used in a model calculation.
     * </p>
     * <br>Sample code:
     * <pre><code>
     *      def ctx = api.getDatamartContext()
     * 	    def ds = ctx.getDataSource("txDS")
     * 	    def t1 = ctx.newQuery(ds)
     * 	    t1.select("sku", "P")
     * 	    t1.select("SUM(amount)", "R")
     * 	    def query = ctx.newSqlQuery()
     * 	    query.addSource(t1)
     * 	    query.setQuery('SELECT P, R::numeric as R FROM T1 ORDER BY P')
     * 	    // add table using a query definition
     * 	    model.addTable('MyTable', query)
     * 	    // add table using fields definitions
     * 	    model.addTable("RevenueByProduct", [
     *               [ name:"productId", label:"Product ID", type:"TEXT", key:true],
     *               [ name:"productGroup", label:"ProductGroup", type:"TEXT", dimension:true],
     *               [ name:"revenue", label:"Revenue", type:"MONEY"]
     *      ])
     * 	    // add table using full definitions
     * 	    model.addTable("RevenueByProduct", [
     * 	            label: "Revenue By Product",
     *              fields: [
     *                  [ name:"productId", label:"Product ID", type:"TEXT", key:true],
     *                  [ name:"productGroup", label:"ProductGroup", type:"TEXT", dimension:true],
     *                  [ name:"revenue", label:"Revenue", type:"MONEY"],
     *              ]
     *      ])
     * </code></pre>
     *
     * @param tableName The name of the table, unique over all tables associated to this model.
     * @param source    One of the following types
     *                  <br><ul>
     *                  <li>{@link net.pricefx.formulaengine.DatamartContext.Query}</li>
     *                  <li>{@link net.pricefx.formulaengine.DatamartContext.SqlQuery} </li>
     *                  <li>{@link Collection} of field definitions as {@link Map}s conforming to {@link DMField}</li>
     *                  <li>Table definition as a {@link Map} conforming to {@link net.pricefx.domain.DMTable}</li>
     *                  </ul>
     *                  <b>Note</b>: For table field types, please refer to {@link net.pricefx.common.api.pa.FieldType Enum FieldType}<p>.
     *
     * @return A {@link DatamartContext.Table} object that can be used as a source to a query in the {@link DatamartContext}.
     * It is the same {@link DatamartContext.Table} object that can be retrieved using {@link #getTable(String)}.
     * @throws Exception in case of error.
     */
    DatamartContext.Table addTable(String tableName, Object source) throws Exception;

    /**
     * Load data into an existing table without truncating it.
     * <br>The the projections of the query passed as a source must match the fields of the table, and the result of
     * the query is used to populate the table.
     * <br>Note that when directly using a {@link net.pricefx.formulaengine.DatamartContext.SqlQuery}, the types of the selected fields
     * must be made explicit via casting for the resulting table to have the proper types.
     * <p>
     * Note: Can only be used in a model calculation.
     * </p>
     * <br>Sample code:
     * <pre><code>
     *      def ctx = api.getDatamartContext()
     * 	    def ds = ctx.getDataSource("txDS")
     * 	    def t1 = ctx.newQuery(ds)
     * 	    t1.select("sku", "P")
     * 	    t1.select("SUM(amount)", "R")
     * 	    def query = ctx.newSqlQuery()
     * 	    query.addSource(t1)
     * 	    query.setQuery('SELECT P, R::numeric as R FROM T1 ORDER BY P')
     * 	    // load table using a query definition
     * 	    def table = model.getTable('MyTable')
     * 	    model.loadTable(table, query)
     * </code></pre>
     *
     * @param table  The table object, see {@link #getTable(String)} or {@link #addTable(String, Object)} to retrieve it.
     * @param source One of the following query types
     *               <br><ul>
     *               <li>{@link net.pricefx.formulaengine.DatamartContext.Query}</li>
     *               <li>{@link net.pricefx.formulaengine.DatamartContext.SqlQuery} </li>
     *               </ul>
     * @return The number of row added.
     * @throws Exception in case of error.
     */
    long loadTable(DatamartContext.Table table, Object source) throws Exception;

    /**
     * Load data into an existing table, optionally truncating it.
     * <br>The the projections of the query passed as a source must match the fields of the table, and the result of
     * the query is used to populate the table.
     * <br>Note that when directly using a {@link net.pricefx.formulaengine.DatamartContext.SqlQuery}, the types of the selected fields
     * must be made explicit via casting for the resulting table to have the proper types.
     * <p>
     * Note: Can only be used in a model calculation.
     * </p>
     * <br>Sample code:
     * <pre><code>
     *      def ctx = api.getDatamartContext()
     * 	    def ds = ctx.getDataSource("txDS")
     * 	    def t1 = ctx.newQuery(ds)
     * 	    t1.select("sku", "P")
     * 	    t1.select("SUM(amount)", "R")
     * 	    def query = ctx.newSqlQuery()
     * 	    query.addSource(t1)
     * 	    query.setQuery('SELECT P, R::numeric as R FROM T1 ORDER BY P')
     * 	    // load table using a query definition
     * 	    def table = model.getTable('MyTable')
     * 	    model.loadTable(table, query, true)
     * </code></pre>
     *
     * @param table    The table object, see {@link #getTable(String)} or {@link #addTable(String, Object)} to retrieve it.
     * @param source   One of the following query types
     *                 <br><ul>
     *                 <li>{@link net.pricefx.formulaengine.DatamartContext.Query}</li>
     *                 <li>{@link net.pricefx.formulaengine.DatamartContext.SqlQuery} </li>
     *                 </ul>
     * @param truncate If the table should be truncated before loading the data.
     * @return The number of row added.
     * @throws Exception in case of error.
     */
    long loadTable(DatamartContext.Table table, Object source, boolean truncate) throws Exception;

    /**
     * Gets a table object representing the DMTable with the given name, which was previously created using the
     * {@link #addTable(String, Object)} method while calculating the model.
     * A reference to this table can then be used when building a {@link DatamartContext.Query} on that DMTable.
     *
     * @param tableName The table name.
     * @return The Table object with the given tableName if it exists, or null if not.
     */
    DatamartContext.Table getTable(String tableName);

    /**
     * Drops the table, uniquely identified by tableName within this model.
     * <p>
     * Note: Can only be used in a model calculation.
     * </p>
     * <br>Sample code:
     * <pre><code>
     * 	    model.dropTable('MyTable')
     * </code></pre>
     *
     * @param tableName The table name.
     * @return True if that table existed, false otherwise.
     */
    boolean dropTable(String tableName);

    /**
     * Truncate the table.
     * <p>
     * Note: Can only be used in a model calculation.
     * </p>
     * <br>Sample code:
     * <pre><code>
     *      def table = model.getTable('MyTable')
     *      model.truncateTable(table)
     * </code></pre>
     *
     * @param table The table.
     */
    void truncateTable(DatamartContext.Table table);

    /**
     * Representation of a node in a model tree (see for ex. {@link #addTree(String, Collection)}
     * The <code>id</code> and <code>parentId</code> are ids of FilterNodes returned by for example {@link #findTreeNode(String, List)}
     */
    class TreeNode {
        public Long id;
        public Long parentId;
        public String name;
        public Filter filter;
        public String output;
        public Collection<TreeNode> children;

        @Override
        public String toString() {
            return "[id:" + id + ", filter: " + filter + ", output:" + output + "]";
        }
    }

}
