package net.pricefx.formulaengine.optimization;

import net.pricefx.domain.DMTable;
import net.pricefx.formulaengine.DatamartContext;

import java.util.List;
import java.util.Map;

public interface ModelFormulaContextReadWrite extends ModelFormulaContextReadOnly {
    /**
     * Add a table in the model. If the table already exists, it is replaced by this
     * new table.
     * <p>
     * A {@link Map} respecting the structure of {@link DMTable}, including:
     * <ul>
     * <li>{@code label}: the table label</li>
     * <li>{@code fields}: a {@link List} of {@link Map} respecting the structure of
     * {@link DMField}, including {@code name}, {@code label}, {@code type},
     * {@code format}, {@code key}, {@code dimension}</li>
     * </ul>
     * <p>
     * Sample code:
     * <pre><code>
     *     model.addTable("RevenueByProduct", [
     *             label: "Revenue By Product",
     *             // the fields must respect DMField structure
     *             fields: [
     *                 [ name:"productId", label:"Product ID", type:"TEXT", key:true],
     *                 [ name:"productGroup", label:"ProductGroup", type:"TEXT", dimension:true],
     *                 [ name:"revenue", label:"Revenue", type:"MONEY"],
     *             ]
     *     ])
     * </code></pre>
     *
     * @param name       The table name to be created in this model
     * @param tableSpecs The definition of the table to be created, see above.
     * @return The created model table
     */
    DatamartContext.Table addTable(String name, Map<String, Object> tableSpecs);

    /**
     * Sample code:
     * <pre><code>
     *     def ctx = api.getDatamartContext()
     *     def query = ctx.newQuery(ctx.getDataSource("txDS"))
     *         .select("sku", "P")
     *         .select("SUM(amount)", "R")
     *     def table = model.getTable('MyTable')
     *     def nbRowAdded = model.loadTable(table, query, false)
     * </code></pre>
     *
     * @param table    A table in this model
     * @param source   The query to load into the table
     * @param truncate If the table should be truncated before loading
     * @return The number of row added
     */
    long loadTable(DatamartContext.Table table, DatamartContext.Query source, boolean truncate);

    /**
     * Sample code:
     * <pre><code>
     *     def ctx = api.getDatamartContext()
     *     def t1 = ctx.newQuery(ctx.getDataSource("txDS"))
     *         .select("sku", "P")
     *         .select("SUM(amount)", "R")
     *     def query = ctx.newSqlQuery()
     *     query.addSource(t1)
     *     query.setQuery('SELECT P, R::numeric as R FROM T1 ORDER BY P')
     *     def table = model.getTable('MyTable')
     *     def nbRowAdded = model.loadTable(table, query, false)
     * </code></pre>
     *
     * @param table    A table in this model
     * @param source   The query to load into the table
     * @param truncate If the table should be truncated before loading
     * @return The number of row added
     */
    long loadTable(DatamartContext.Table table, DatamartContext.SqlQuery source, boolean truncate);

    /**
     * @param table A table in the model
     */
    void truncateTable(final DatamartContext.Table table);

    /**
     * @param table A table in the model
     */
    void dropTable(final DatamartContext.Table table);

    /**
     * It is the responsibility of the caller to serialize {@code parameters} however they are required by the image.
     * <p>
     * {@code jobName} should be unique inside the step and model formula calculation that triggers them.
     * Since Job Trigger Calculations are created from within an existing formula calculation, its full name will be
     * constructed from the calculation name and the job name passed to this method: {@code calculationName:jobName}.
     * This name can be used to access the Job Trigger Calculation results with {@link #outputs(String, String)}.
     * <p>
     * Multiple job trigger calculation in a given formula calculation will run in parallel.
     *
     * @param imageName  A docker image name, without the tag
     * @param imageTag   A tag for the docker image name
     * @param parameters Some parameters understood by the image used
     * @param jobName    A name for the job, unique per current step and calculation
     */
    void startJobTriggerCalculation(String imageName, String imageTag, String parameters, String jobName);

    /**
     * @return a context for read/write operations on the lookup tables of this model
     */
    @Override
    ModelLookupTablesContextReadWrite lookupTablesContext();

    /**
     * @return a context for read/write operations on the filter trees of this model
     */
    @Override
    ModelFilterTreesContextReadWrite filterTreesContext();
}
