package net.pricefx.formulaengine.scripting.inputbuilder;

import net.pricefx.common.api.InputType;
import net.pricefx.common.api.PriceFxInterface;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * Builder class for input type Option and Options
 * ({@link InputType#OPTION} and {@link InputType#OPTIONS}).
 */
public class OptionInputBuilder extends AbstractInputBuilder<OptionInputBuilder, Object> {

    private List<String> options;

    public OptionInputBuilder(InputBuilderContext builderContext, InputType type, String paramName) {
        super(builderContext, type, paramName);
    }

    /**
     * Add options to the OptionInputBuilder instance.
     * <p>Example:
     * <blockquote><pre>
     * Map driverClasses = ...
     *
     * optionInputBuilder.addOptions(driverClasses?.driverClassLabel as List)
     * </pre></blockquote>
     *
     * @param options List of options names
     * @return self
     */
    public OptionInputBuilder setOptions(List<String> options) {
        this.options = options;
        cp.setValueOptions(options);
        return this;
    }

    /**
     * Add option to the OptionInputBuilder instance.
     * <p>Example:
     * <blockquote><pre>
     * Map driverClasses = ...
     *
     * driverClasses.each { String driverClassName, String driverClassLabel -&gt;
     *      optionInputBuilder.addOption(driverClassLabel)
     * }
     * </pre></blockquote>
     *
     * @param option option name
     * @return self
     */
    public OptionInputBuilder addOption(String option) {
        if (options == null) {
            options = new ArrayList<>();
        }
        options.add(option);
        cp.setValueOptions(options);
        return this;
    }

    /**
     * Set labels of the OptionInputBuilder instance.
     * <p>Example:
     * <blockquote><pre>
     * OptionInputBuilder builder = api.inputBuilderFactory()
     *                                 .createRadioEntry(inputName)
     *                                 .setLabels([model : "Services Model",
     *                                             type  : "Service Type"])
     *                                 .setValue(inputValue)
     * </pre></blockquote>
     *
     * @param labels Map of labs to be set
     * @return self
     */
    public OptionInputBuilder setLabels(Map<String, Object> labels) {
        if (cp != null && labels != null && !labels.isEmpty()) {
            cp.setConfigParameter(PriceFxInterface.FIELD_CONTEXTPARAM_CONFIG_LABELS, labels);
        }
        return this;
    }

}
