package net.pricefx.formulaengine.scripting.portlets;

import java.io.Serializable;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.googlecode.genericdao.search.Filter;

import net.pricefx.common.api.*;
import net.pricefx.domain.LookupTable;
import net.pricefx.domain.PersistedObject;
import net.pricefx.domain.PriceGrid;
import net.pricefx.domain.Pricelist;
import net.pricefx.formulaengine.scripting.AppPages;
import net.pricefx.formulaengine.scripting.Matrix2D;
import net.pricefx.server.dto.calculation.ComplexCalculationResult;


/**
 * This class is used to transport a calculated matrix-style result
 * It is basically a list of maps with the addition that the column names are clearly named and can have a format
 * This means:
 * - the UI knows exactly which columns to display (even if some maps do not have those as keys)
 * - the entry maps can have more "columns" (or keys) than those that are displayed; maybe useful for calculation
 *
 * @author cat
 *
 */
public interface ResultMatrix extends ComplexCalculationResult, AppPages, Portlet {

    /**
     * Gets the current title
     *
     * @return The current title
     */
    String getTitle();

    /**
     * Sets the title that will be shown as window title of the result matrix in the client
     *
     * @param title The title
     */
    default void setTitle(String title) {
        withTitle(title);
    }

    /**
     * Sets the title that will be shown as window title of the result matrix in the client
     *
     * @param title The title
     */
    ResultMatrix withTitle(String title);

    /**
     * Gets the preference name
     *
     * @return The preference name
     */
    String getPreferenceName(); // TODO: do we need this in the public API? It's more for JSON serialization

    /**
     * Tells whether saving of matrix grid preferences is disabled.
     *
     * @return true if saving of grid preferences is disabled. Default is false
     * @since 7.2 Bijou PR2
     */
    boolean isPreferencesDisabled();

    /**
     * Allows to disable saving preferences of the matrix grid (columns, sorting, filtering). If not disabled, the
     * preferences may be be saved with the global dashboard preferences or attached to a specific preferences name
     * defined with setPreferenceName()/withPreferenceName()
     *
     * @param disablePreferences true means saving grid preferences is disabled. Default is false
     * @since 7.2 Bijou PR2
     */
    default void setDisablePreferences(boolean disablePreferences) {
        withDisablePreferences(disablePreferences);
    }

    /**
     * Allows to disable saving preferences of the matrix grid (columns, sorting, filtering). If not disabled, the
     * preferences may be be saved with the global dashboard preferences or attached to a specific preferences name
     * defined with setPreferenceName()/withPreferenceName()
     *
     * @param disablePreferences true means saving grid preferences is disabled. Default is false
     * @since 7.2 Bijou PR2
     */
    ResultMatrix withDisablePreferences(boolean disablePreferences);

    /**
     * UI sorting of matrix grid
     *
     * @return true if sorting is disabled on all columns. Default is false
     */
    boolean isSortingDisabled();

    /**
     * Set sorting capability in UI grid
     *
     * @param disableSorting true: sorting is disabled in UI on all columns
     */
    default void setDisableSorting(boolean disableSorting) {
        withDisableSorting(disableSorting);
    }

    /**
     * Set sorting capability in UI grid
     *
     * @param disableSorting true: sorting is disabled in UI on all columns
     */
    ResultMatrix withDisableSorting(boolean disableSorting);

    /**
     * Is HTML escaping enabled?
     *
     * @return true if the matrix should display raw text as is (i.e. not try to interpret it as HTML)
     */
    boolean isEscapeHTML();

    /**
     * Set HTML escaping flag. Default: false (i.e. values will be interpreted as HTML)
     *
     * @param escapeHTML
     */
    default void setEscapeHTML(boolean escapeHTML) {
        withEscapeHTML(escapeHTML);
    }

    /**
     * Set HTML escaping flag. Default: false (i.e. values will be interpreted as HTML)
     *
     * @param escapeHTML
     */
    ResultMatrix withEscapeHTML(boolean escapeHTML);

    /**
     * Sets the so-called preference name. I.e. all result matrixes with the same preference name will share the same preference setting.
     * The preference setting can be managed as usual in the client and can be set as global to predefine a good setting
     *
     * @param preferenceName The preference name
     */
    default void setPreferenceName(String preferenceName) {
        withPreferenceName(preferenceName);
    }

    /**
     * Sets the so-called preference name. I.e. all result matrixes with the same preference name will share the same preference setting.
     * The preference setting can be managed as usual in the client and can be set as global to predefine a good setting
     *
     * @param preferenceName The preference name
     */
    ResultMatrix withPreferenceName(String preferenceName);

    /**
     * Adds translated labels for the given column.
     *
     * @param column       The column name (as defined e.g. in the constructor)
     * @param translations A map with the translations. The key of the map is the locale, like "de".
     */
    default void addColumnTranslation(String column, Map<String, String> translations) {
        withColumnTranslation(column, translations);
    }

    /**
     * Adds translated labels for the given column.
     *
     * @param column       The column name (as defined e.g. in the constructor)
     * @param translations A map with the translations. The key of the map is the locale, like "de".
     */
    ResultMatrix withColumnTranslation(String column, Map<String, String> translations);

    /**
     * Gets current column translations in raw format
     *
     * @return
     */
    Map<String, String> getColumnTranslations(); // TODO: do we need this in the public API? It's more for JSON serialization

    /**
     * Adds a column to the existing set of columns
     *
     * @param column The columns name
     */
    default void addColumn(String column) {
        withColumn(column);
    }

    /**
     * Adds a column to the existing set of columns
     *
     * @param column The columns name
     */
    ResultMatrix withColumn(String column);

    /**
     * Adds a column to the existing set of columns including a tooltip
     *
     * @param column The columns name
     */
    default void addColumn(String column, String tooltip) {
        withColumn(column, tooltip);
    }

    /**
     * Adds a column to the existing set of columns including a tooltip
     *
     * @param column The columns name
     */
    ResultMatrix withColumn(String column, String tooltip);

    /**
     * Adds columns to the existing set of columns
     *
     * @param columns Collection of columns names
     */
    default void addColumns(Collection<String> columns) {
        withColumns(columns);
    }

    /**
     * Adds columns to the existing set of columns
     *
     * @param columns Collection of columns names
     */
    ResultMatrix withColumns(Collection<String> columns);

    /**
     * Adds a data row. The keys of the row-map should match the column names defined earlier
     *
     * @param row
     */
    void addRow(Map<String, Object> row);

    /**
     * Adds a data row. The order of the values is assumed to be the same as the matrix column order.
     *
     * @param row Row as List
     */
    default void addRow(List<Object> row) {
        withRow(row);
    }

    /**
     * Adds a data row. The order of the values is assumed to be the same as the matrix column order.
     *
     * @param row Row as List
     */
    ResultMatrix withRow(List<Object> row);

    /**
     * Adds multiple data rows. The keys of the row-map should match the column names defined earlier
     *
     * @param rows Row as List of Maps
     */
    ResultMatrix withRows(List<Map<String, Object>> rows);

    /**
     * Adds a data row. The order of the values is assumed to be the same as the matrix column order.
     *
     * @param values Row as Array
     */
    void addRow(Object... values);

    /**
     * Gets a list of defined columns
     *
     * @return The columns
     */
    List<String> getColumns(); // TODO: do we need this in the public API? It's more for JSON serialization

    /**
     * Returns the current list entries as a list (=the rows) of maps (= the row content)
     *
     * @return List of maps
     */
    List<Map<String, Object>> getEntries(); // TODO: do we need this in the public API? It's more for JSON serialization

    /**
     * Retrieves a particular row based on a value check
     *
     * @param key   The column name to search on
     * @param value The value to match the row's column value to
     * @return The row
     */
    Map<String, Object> findRow(String key, Object value); // TODO: do we need this in the API? ResultMatrix is for UI, not data manipulation

    /**
     * Transpose the matrix.
     * <p>
     * <br/><br/>
     * <p>
     * Example:
     * <PRE>
     * ProductGroup | AvgPrice | AvgMargin &lt;- header row
     * -----------------------------------
     * Beef     |   4.1€   |   3.6%
     * Pork     |   2.2€   |   1.1%
     * Lamb     |   4.6€   |   5.2%
     * </PRE>
     * <p>
     * Transposed matrix:
     * <PRE>
     * transpose("ProductGroup")
     * <p>
     * ProductGroup | Beef | Pork | Lamb &lt;- header row
     * ---------------------------------
     * AvgPrice   | 4.1€ | 2.2€ | 4.6€
     * AvgMargin  | 3.6% | 1.1% | 5.2%
     * </PRE>
     *
     * @param newHeaderColumnName Name of the column that will become the new header row
     * @return The transposed ResultMatrix (new instance)
     */
    default ResultMatrix transpose(String newHeaderColumnName) {
        throw new RuntimeException("Method ResultMatrix.transpose() is not supported by debugger");
    }

    /**
     * Transpose the matrix.
     * <p>
     * <br/><br/>
     * <p>
     * Example:
     * <PRE>
     * ProductGroup | AvgPrice | AvgMargin  &lt;- header row
     * ------------------------------------
     * Beef     |   4.1€   |   3.6%
     * Pork     |   2.2€   |   1.1%
     * Lamb     |   4.6€   |   5.2%
     * </PRE>
     * <p>
     * Transposed matrix:
     * <PRE>
     * transpose("ProductGroup", "PG Label")
     * <p>
     * Beef | Pork | Lamb |   PG Label    &lt;- header row
     * -----------------------------------
     * Beef | Pork | Lamb | ProductGroup
     * 4.1€ | 2.2€ | 4.6€ |   AvgPrice
     * 3.6% | 1.1% | 5.2% |   AvgMargin
     * </PRE>
     *
     * @param newHeaderColumnName         Name of the column that will become the new header row
     * @param oldColLabelsInNewColumnName Name of the new column that will receive the old column headers as values (optional)
     * @return The transposed ResultMatrix (new instance)
     */
    default ResultMatrix transpose(String newHeaderColumnName, String oldColLabelsInNewColumnName) {
        throw new RuntimeException("Method ResultMatrix.transpose() is not supported by debugger");
    }

    default Matrix2D toMatrix2D() {
        throw new RuntimeException("Method ResultMatrix.transpose() is not supported by debugger");
    }

    /**
     * Sets a default column format
     *
     * @param fft
     */
    default void setDefaultFormat(FieldFormatType fft) {
        withDefaultFormat(fft);
    }

    /**
     * Sets a default column format
     *
     * @param fft
     */
    ResultMatrix withDefaultFormat(FieldFormatType fft);

    /**
     * Gets the default format
     *
     * @return Default format
     */
    FieldFormatType getDefaultFormat();

    /**
     * Gets individual column formats (if specified)
     *
     * @return A Map with column name and format
     */
    Map<String, FieldFormatType> getColumnFormats();

    /**
     * Sets a column format for a specified column
     *
     * @param column The column name
     * @param fft    The format
     */
    default void setColumnFormat(String column, FieldFormatType fft) {
        withColumnFormat(column, fft);
    }

    /**
     * Sets a column format for a specified column
     *
     * @param column The column name
     * @param fft    The format
     */
    ResultMatrix withColumnFormat(String column, FieldFormatType fft);

    /**
     * Sets a column formats for multiple columns. Adds column if not existing yet
     *
     * @param columnFormats A key value map of column:format type
     */
    default ResultMatrix withColumnFormats(Map<String, FieldFormatType> columnFormats) {
        if (columnFormats != null) {
            for (String key : columnFormats.keySet()) {
                addColumn(key);
                setColumnFormat(key, columnFormats.get(key));
            }
        }
        return this;
    }

    /**
     * Gets the currently assigned column tooltips
     *
     * @return
     */
    Map<String, String> getColumnTooltips(); // TODO: do we need this in the public API? It's more for JSON serialization

    /**
     * Adds a tooltip to a column
     *
     * @param column  The column name
     * @param tooltip The tooltip text
     */
    default void setColumnTooltip(String column, String tooltip) {
        withColumnTooltip(column, tooltip);
    }

    /**
     * Adds a tooltip to a column
     *
     * @param column  The column name
     * @param tooltip The tooltip text
     */
    ResultMatrix withColumnTooltip(String column, String tooltip);

    boolean isEnableClientFilter(); // TODO: do we need this in the public API? It's more for JSON serialization

    default void setEnableClientFilter(boolean enableClientFilter) {
        withEnableClientFilter(enableClientFilter);
    }

    ResultMatrix withEnableClientFilter(boolean enableClientFilter);

    /**
     * Creates a styled cell that can then be added as row value
     *
     * @param value The value of the cell
     * @return The styled cell object that can then be further styled with e.g. the flow pattern .withXXX() methods
     */
    StyledCell styledCell(Object value);

    /**
     * Creates a styled cell that can then be added as row value
     *
     * @param value     The value of the cell
     * @param textColor The text color of the cell
     * @param bgColor   The background color of the cell (HTML hex value)
     * @return The styled cell object
     */
    StyledCell styledCell(Object value, String textColor, String bgColor);

    /**
     * Creates a styled cell that can then be added as row value
     *
     * @param value     The value of the cell
     * @param textColor The text color of the cell
     * @param bgColor   The background color of the cell (HTML hex value)
     * @param weight    The font weight
     * @return The styled cell object
     */
    StyledCell styledCell(Object value, String textColor, String bgColor, String weight);

    /**
     * Creates a styled cell that can then be added as row value
     *
     * @param value     The value of the cell
     * @param textColor The text color of the cell
     * @param bgColor   The background color of the cell (HTML hex value)
     * @param weight    The font weight
     * @param alignment Cell alignement value (valid CSS)
     * @return The styled cell object
     */
    StyledCell styledCell(Object value, String textColor, String bgColor, String weight, String alignment);

    /**
     * Creates an image cell that can then be added as row value. Value should be a valid image URL
     *
     * @param value The image URL
     * @return The image cell object
     */
    ImageCell imageCell(Object value);

    ImageCell libraryImage(String mainName, String statusName);

    default LinkCell linkToPriceGrid(Object value, long priceGridID, Filter filter) {
        var state = new HashMap<String, Object>(2);
        state.put(PriceFxInterface.FIELD_ID,
                priceGridID + "." + PriceGrid.class.getAnnotation(PersistedObject.PersistedObjectMeta.class).typeCode());
        state.put(PriceFxInterface.PARAM_FILTER_CRITERIA, filter);

        return linkCell(value, PageTokens.PRICE_GRID_PAGE, state);
    }

    default LinkCell linkToPriceList(Object value, long priceListID, Filter filter) {
        var state = new HashMap<String, Object>(2);
        state.put(PriceFxInterface.FIELD_ID,
                priceListID + "." + Pricelist.class.getAnnotation(PersistedObject.PersistedObjectMeta.class).typeCode());
        state.put(PriceFxInterface.PARAM_FILTER_CRITERIA, filter);

        return linkCell(value, PageTokens.PRICE_LIST_PAGE, state);
    }

    default LinkCell linkToPriceParameter(Object value, long priceParameterID) {
        var state = new HashMap<String, Object>(2);
        state.put(PriceFxInterface.FIELD_ID,
                priceParameterID + "." + LookupTable.class.getAnnotation(PersistedObject.PersistedObjectMeta.class).typeCode());

        return linkCell(value, PageTokens.PRICING_PARAMETERS_PAGE, state);
    }

    LinkCell linkCell(final Object value, final String targetPage, final Object targetPageState);

    /**
     * Displays a button inside a cell of the table.
     * <br/>
     * <p>Simple button example:</p>
     * <blockquote>
     * <pre>
     *     def targetPageState = [quoteType: 'Quote with Inputs in Header', Customer: 'custId']
     *     def cell = matrix.linkButtonCell('New Quote', AppPages.QC_NEW_QUOTE, targetPageState, null)
     *     matrix.addRow('Customer A', 'Value X', cell)
     * </pre>
     * </blockquote>
     *
     * <p>Advanced button example:</p>
     * <blockquote>
     * <pre>
     *     def customerPageState = [Customer: 'custId']
     *     def groupPageState = [customerFieldName:'customerId', customerFieldValue:custId]
     *     def builder = api.inputBuilderFactory()
     *                      .createQuoteType("New quote")
     *                      .buildContextParameter()
     *     def cell = matrix.linkButtonCell('cell value' &sol;*not displayed*&sol;, AppPages.QC_NEW_QUOTE, groupPageState, builder)
     *     matrix.addRow('Customer A', 'Value X', cell)
     * </pre>
     * </blockquote>
     *
     * <p>Note: This is available in Unity UI only. No backporting planned for the Classic UI.</p>
     *
     * @since 9.0 Hurricane
     */
    LinkButtonCell linkButtonCell(final Object value, final String targetPage, final Object targetPageState,
                                  final Object buttonState);

    BackEndCell backEndAction(String label, String url, String payload, String successMessage, String failMessage);

    /**
     * <p>Allows UI users to select multiple rows and adds an action button next to the ResultMatrix. Multiple such
     * action buttons can be defined, identified by the provided <code>id String</code>.</p>
     * <p>The action will be handled by a dedicated logic, which receives an <code>InputMatrix</code> parameter
     * named after the provided <code>id</code> and featuring the row values of a subset of the columns.</p>
     * <p>Example:</p>
     * <pre>
     *     def matrix = api.newMatrix("col1", "col2", "col3")
     *     matrix.addRow(1, "a", "f")
     *     matrix.addRow(2, "b", "g")
     *     matrix.rowSelectionBackEndAction("actionId")
     *                 .withLogicName("action_handler")
     *                 .withColumns("col2", "col3")
     *                 .withButtonLabel("Click here")
     *                 .withSuccessMessage("Success")
     *                 .withFailureMessage("Sorry")
     *
     * </pre>
     *
     * <p>Note: This is available in Unity UI only. No backporting planned for the Classic UI.</p>
     * <p>
     * See {@link BackEndAction} to explore the full list of methods allowing to configure an action.
     *
     * @param id a <code>String</code> identifying the action
     * @return a customizable object representing an action button, which the UI will then render and make functional
     * @since 7.0 Bijou
     */
    BackEndAction rowSelectionBackEndAction(final String id);

    DownloadCell downloadButton(String label, String url, String payload);

    Cells cells(String label, Object... cells);

    /**
     * Configures a client-side UI event. Use case is usually a refresh of some sort when a result matrix row is selected
     * Example code:
     * <pre>
     * matrix.onRowSelection().triggerEvent(api.dashboardWideEvent("CustomerIdChanged-ED")).withColValueAsEventDataAttr("CustomerID", "customerId")
     * </pre>
     * The embedded dashboard that receives the event could then look like:
     * <pre>
     *  return api.dashboard("EmbeddedDashboard")
     *
     *    // Here the user can assign values to the embedded
     *    // dashboard inputs common to all three modes
     *    .setParam("Customer Id", api.input("Customer Id"))
     *    .setParam("Year", api.input("Year"))
     *
     *    // Show the dashboard embedded ...
     *    .showEmbedded()
     *    // ... and reevaluate it on "CustomerIdChanged-ED"
     *    // Note: "dashboardWideEvent()" fce makes the event local
     *    // to the containing dashboard instance
     *    .andRecalculateOn(api.dashboardWideEvent("CustomerIdChanged-ED"))
     *    // Pull the "customerId" attribute out of the event payload and expose it
     *    // as the "Customer Id" input to the embedded dashboard
     *    .withEventDataAttr("customerId").asParam("Customer Id")
     * </pre>
     *
     * @return The selection event to further configure it in a flow pattern
     */
    OnRowSelection onRowSelection();

    @Override
    default String getResultType() {
        return CalculationResultType.MATRIX.name();
    }


    void addFormatsToResultMatrix();

    /**
     * Search for non null {@link FieldFormatType#DATETIME}</a> columns and convert them into
     * {@link org.joda.time.DateTime}</a>
     */
    void normalizeDateTimes();


    interface Cell {
    }

    interface StyledCell extends Cell {

        /**
         * Set the foreground color of the value font
         *
         * @param color
         * @return
         */
        StyledCell withColor(String color);

        /**
         * Set the background color of the cell
         *
         * @param bgColor
         * @return
         */
        StyledCell withBackground(String bgColor);

        /**
         * Set the font weight (in standard CSS terms) of the value text
         *
         * @param fontWeight
         * @return
         */
        StyledCell withFontWeight(String fontWeight);

        /**
         * Specify the alignment of the cell
         *
         * @param alignment
         * @return
         */
        StyledCell withAlignment(String alignment);

        /**
         * Show a tooltip text in the cell
         *
         * @param tooltip
         * @return
         */
        StyledCell withToolTip(String tooltip);

        /**
         * Apply any other valid CSS instructions to the cell
         *
         * @param extraCSS
         * @return
         */
        StyledCell withCSS(String extraCSS);
    }

    interface LinkCell extends Cell {
    }

    interface LinkButtonCell extends LinkCell {
    }

    /**
     * Create a link/button that should do some desired action to pricefx backend (unity only)
     * [url:"pricelistmanager.revoke..../...", payload: [... : ...], successMessage: "hey, it is ok!", failMessage: "ayay, this failed"]
     */
    interface BackEndCell extends Cell {
    }

    /**
     * Create a link/button that should do initiate download from pricefx backend (unity only)
     * [url:"pricelistmanager.fetchpdf..../...", payload: [... : ...]]
     */
    interface DownloadCell extends Cell {
    }

    /**
     * List of other cells (to be shown as dropdown inline)
     */
    interface Cells extends Cell {
    }

    interface ImageCell extends Cell {
    }

    interface OnRowSelection {

        TriggerEvent triggerEvent(String eventName);

        interface TriggerEvent {

            TriggerEvent withColValueAsEventDataAttr(String colName, String attrName);
        }
    }

    /**
     * <p>Object representing a button that handles multi-row actions.</p>
     * <p>The action will be handled by a dedicated logic, which receives an <code>InputMatrix</code> parameter
     * named after the provided <code>id String</code> and featuring the row values of a subset of the columns.</p>
     *
     * <p>Note: This is available in Unity UI only. No backporting planned for the Classic UI.</p>
     * <p>
     * See {@link ResultMatrix#rowSelectionBackEndAction(String)} to find out how to create an instance of
     * <code>ResultMatrixBackEndAction</code>.
     *
     * @since 7.0 Bijou
     */
    interface BackEndAction extends Serializable {

        /**
         * @param name the logic to execute
         * @since 7.0 Bijou
         */
        BackEndAction withLogicName(final String name);

        /**
         * @param date <b>optional</b> targetDate in case there are multiple active versions with the same name
         * @since 7.0 Bijou
         */
        BackEndAction withTargetDate(final Date date);

        /**
         * @param names list of column names that will be passed as <code>InputMatrix</code> to the logic
         *              (replaces any previously set column names)
         * @since 7.0 Bijou
         */
        BackEndAction withColumns(final String... names);

        /**
         * @param name name to add in the list of columns that will be passed as <code>InputMatrix</code> to the logic
         * @since 7.0 Bijou
         */
        BackEndAction addColumn(final String name);

        /**
         * @param name  name of the input parameter that will be passed to the logic
         * @param value value of the input parameter that will be passed to the logic (keep in mind that this value makes
         *              a round trip to the front end in JSON format)
         * @since 7.0 Bijou
         */
        BackEndAction addFormulaInput(final String name, final Object value);

        /**
         * @param label the label of the action button
         * @since 7.0 Bijou
         */
        BackEndAction withButtonLabel(final String label);

        /**
         * @param tooltip <b>optional</b> hover message for additional info about the action
         * @since 7.0 Bijou
         */
        BackEndAction withButtonTooltip(final String tooltip);

        /**
         * @param text message to display in case of success of the backend call
         * @since 7.0 Bijou
         */
        BackEndAction withSuccessMessage(final String text);

        /**
         * @param text message to display in case of failure of the backend call
         * @since 7.0 Bijou
         */
        BackEndAction withFailureMessage(final String text);

    }

    // BACK COMPATIBILITY STUFF - see SandboxApi_newMatrixIT."ResultMatrix refactoring is back compatible"
    
    // no matter what ResultMatrix import statement (like in the IT), the classes/interface whitelisted in
    // the sandbox take precedence, so that's where the back-compatibility inner interfaces need to be

    /**
     * @deprecated better use the shorter {@link Cell}
     */
    @Deprecated
    interface ResultMatrixCell extends Cell {
    }

    /**
     * @deprecated better use the shorter {@link StyledCell}
     */
    @Deprecated
    interface ResultMatrixStyledCell extends StyledCell {
    }

    /**
     * @deprecated better use the shorter {@link LinkCell}
     */
    @Deprecated
    interface ResultMatrixLinkCell extends LinkCell {
    }

    /**
     * @deprecated better use the shorter {@link LinkButtonCell}
     */
    @Deprecated
    interface ResultMatrixLinkButtonCell extends LinkButtonCell {
    }

    /**
     * @deprecated better use the shorter {@link BackEndCell}
     */
    @Deprecated
    interface ResultMatrixBackEndCell extends BackEndCell {
    }

    /**
     * @deprecated better use the shorter {@link DownloadCell}
     */
    @Deprecated
    interface ResultMatrixDownloadCell extends DownloadCell {
    }

    /**
     * @deprecated better use the shorter {@link Cells}
     */
    @Deprecated
    interface ResultMatrixCells extends Cells {
    }

    /**
     * @deprecated better use the shorter {@link ImageCell}
     */
    @Deprecated
    interface ResultMatrixImageCell extends ImageCell {
    }

    /**
     * @deprecated better use the shorter {@link OnRowSelection}
     */
    @Deprecated
    interface ResultMatrixOnRowSelection extends OnRowSelection {

        @Deprecated
        @Override
        ResultMatrixTriggerEvent triggerEvent(String eventName);

        @Deprecated
        interface ResultMatrixTriggerEvent extends TriggerEvent {

            @Deprecated
            @Override
            ResultMatrixTriggerEvent withColValueAsEventDataAttr(String colName, String attrName);
        }
    }

    /**
     * @deprecated better use the shorter {@link BackEndAction}
     */
    @Deprecated
    interface ResultMatrixBackEndAction extends BackEndAction {
    }

}
