package net.pricefx.server.dto.calculation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import net.pricefx.common.api.CalculationResultType;
import net.pricefx.common.api.PriceFxInterface;
import net.pricefx.formulaengine.scripting.PublicGroovyAPI;
import net.pricefx.formulaengine.scripting.portlets.ResultMatrix;

// TODO replace this by an interface in order to hide unnecessary public methods, especially the constructor
public class ResultHighchart implements ComplexCalculationResult {

	private static final long serialVersionUID = 7030L; // picked Highchart release + 1 digit

	private final Map<String, ?> chartDefinition;

	private final List<String> modulesList = new ArrayList<>();

	private boolean showDataTab = false;

	private List<String> headersList = null;

	private List<ResultMatrix> resultMatrixList = null;

	private OnPointClickImpl onClick = null;

	/**
	 * Not intended for public use. Using this constructor will not fully initialize the instance,
	 * please prefer {@code api.buildHighchart()}.
	 *
	 * @see PublicGroovyAPI#buildHighchart
	 */
	public ResultHighchart(final Map<String, ?> definition) {
		this.chartDefinition = definition;
	}

	/**
	 * Not intended for public use, required for JSON serialization.
	 */
	public Map<String, ?> getChartDefinition() {
		return chartDefinition;
	}

	/**
	 * By default, these modules are already available: boost, drilldown, exporting, export-data, heatmap,
	 * no-data-to-display, treemap. If your chart requires additional modules, they can be specified here.
	 * The application ensures the files load only once.
	 * <p>
	 * Note regarding the "boost" module: The <a href="https://api.highcharts.com/highcharts/boost">Highcharts
	 * documentation</a> states that it is enabled by default, but in Pricefx you will need to explicitly set
	 * the option <code>boost.enabled = true</code>
	 *
	 * @param modules    one or more Highcharts module names, without the path and without the ".js" extension.
	 *                   e.g. "annotations", "funnel", "variwide"
	 * @return the {@link ResultHighchart} instance, for convenience
	 * @see <a href="https://www.highcharts.com/docs">https://www.highcharts.com/docs</a> to find out what modules
	 *      are needed for each chart type.
	 */
	@SuppressWarnings("unchecked")
	public ResultHighchart addModule(String... modules) {
//		modulesList.addAll(Collections.arrayToList(modules));
		return this;
	}

	/**
	 * Not intended for public use, required for JSON serialization.
	 */
	public List<String> getModulesList() {
		return modulesList;
	}

	////////////
	// DATA TAB
	////////////

	/**
	 * Show a tab set, with the chart in the first tab, and data grids in the other tabs.
	 * By default, a single grid will be shown by extracting the data using the native JavaScript
	 * Highcharts API. The headers of this grid can be customized via {@code setDataHeaders()},
	 * or the whole grid can be provided in a {@link ResultMatrix} instance via
	 * {@code setResultMatrixTabs()}.
	 *
	 * Note that for convenience, calling {@code setDataHeaders()}, {@code setResultMatrixTabs()},
	 * or {@code addResultMatrixTab()}
	 * will set this to true.
	 *
	 * @param show    Data tab will be shown if set to true. Default is false.
	 * @return the {@link ResultHighchart} instance, for convenience
	 * @see #setDataHeaders
	 * @see #setResultMatrixTabs
	 * @see #addResultMatrixTab
	 */
	public ResultHighchart setShowDataTab(final boolean show) {
		showDataTab = show;
		return this;
	}

	/**
	 * Short for {@code setShowDataTab(true)}
	 * @return the {@link ResultHighchart} instance, for convenience
	 */
	public ResultHighchart showDataTab() {
		return setShowDataTab(true);
	}

	/**
	 * Not intended for public use, required for JSON serialization.
	 */
	public boolean getShowDataTab() {
		return showDataTab;
	}

	/**
	 * To override the default column headers of the data tab table.
	 *
	 * For convenience, this will also call {@code setShowDataTab(true)}
	 *
	 * @param headers    One or more header labels, left to right
	 * @return the {@link ResultHighchart} instance, for convenience
	 */
	@SuppressWarnings("unchecked")
	public ResultHighchart setDataHeaders(String... headers) {
//		headersList = Collections.arrayToList(headers);
		return showDataTab();
	}

	/**
	 * Not intended for public use, required for JSON serialization.
	 */
	public List<String> getDataHeaders() {
		return headersList;
	}

	/**
	 * The automatically filled data tab can be replaced by customized ResultMatrix data.
	 * It's even possible to add multiple tabs by providing multiple ResultMatrix instances.
	 *
	 * Make sure to set a title via {@code ResultMatrix.setTitle()} as this will be used for the
	 * tab label.
	 *
	 * Note that:
	 * <ul>
	 * <li>This method will overwrite any previously set or added {@code ResultMatrix}</li>
	 * <li>For convenience, this will also call {@code setShowDataTab(true)}</li>
	 * </ul>
	 *
	 * @param matrices    One or more ResultMatrix instances to display in the tab set
	 * @return the {@link ResultHighchart} instance, for convenience
	 * @see #setShowDataTab
	 * @see {@link ResultMatrix#setTitle}
	 */
	@SuppressWarnings("unchecked")
	public ResultHighchart setResultMatrixTabs(ResultMatrix... matrices) {
//		resultMatrixList = Collections.arrayToList(matrices);
		return showDataTab();
	}

	/**
	 * The automatically filled data tab can be replaced by customized ResultMatrix data.
	 * It's even possible to add multiple tabs by providing multiple ResultMatrix instances.
	 *
	 * Make sure to set a title via {@code ResultMatrix.setTitle()} as this will be used for the
	 * tab label.
	 *
	 * Note that:
	 * <ul>
	 * <li>This method will <b>not</b> overwrite previously set or added {@code ResultMatrix}</li>
	 * <li>For convenience, this will also call {@code setShowDataTab(true)}</li>
	 * </ul>
	 *
	 * @param matrix    Add one ResultMatrix instance to display in its own tab
	 * @return the {@link ResultHighchart} instance, for convenience
	 * @see #setShowDataTab
	 * @see {@link ResultMatrix#setTitle}
	 */
	public ResultHighchart addResultMatrixTab(ResultMatrix matrix) {
		if (resultMatrixList == null) {
			resultMatrixList = new ArrayList<>();
		}
		resultMatrixList.add(matrix);
		return showDataTab();
	}

	/**
	 * Not intended for public use, required for JSON serialization.
	 */
	public List<ResultMatrix> getResultMatrixList() {
		return resultMatrixList;
	}

	@Override
	public String getResultType() {
		return CalculationResultType.HIGHCHART.name();
	}

	////////////////
	// CLICK EVENTS
	////////////////

	public interface OnPointClick {

		/**
		 * Define an event that will be triggered when the user click in the chart.
		 *
		 * The name of the event is set as a parameter, the desired payload of the event can
		 * be set using instance methods of the returned {@link TriggerEvent}
		 *
		 * @param eventName the name of the event
		 * @return a new {@link TriggerEvent} instance
		 */
		TriggerEvent triggerEvent(String eventName);

		/**
		 * @return the {@link ResultHighchart} instance, for convenience
		 */
		ResultHighchart back();

	}

	public interface TriggerEvent {

		/**
		 * Short for {@code withCustomEventDataAttr("x")}
		 *
		 * @return the {@link TriggerEvent} instance, for convenience
		 */
		TriggerEvent withXAsEventDataAttr();

		/**
		 * Short for {@code withCustomEventDataAttr("y")}
		 *
		 * @return the {@link TriggerEvent} instance, for convenience
		 */
		TriggerEvent withYAsEventDataAttr();

		/**
		 * Short for {@code withCustomEventDataAttr("name")}
		 *
		 * @return the {@link TriggerEvent} instance, for convenience
		 */
		TriggerEvent withNameAsEventDataAttr();

		/**
		 * Short for {@code withCustomEventDataAttr("category")}
		 *
		 * @return the {@link TriggerEvent} instance, for convenience
		 */
		TriggerEvent withCategoryAsEventDataAttr();

		/**
		 * Pick one or more attributes of the Highchart data Point that should be included
		 * in the triggered event's payload. The attribute can then be mapped to a Dashboard
		 * Input Parameter (see {@link #onPointClick} for sample code).
		 *
		 * @return the {@link TriggerEvent} instance, for convenience
		 */
		TriggerEvent withCustomEventDataAttr(String customAttribute);

		/**
		 * @return the {@link OnPointClick} instance, for convenience
		 */
		OnPointClick back();

	}

	private class OnPointClickImpl implements OnPointClick {

		public static final String EVENT_NAME_KEY = PriceFxInterface.CHARTDEF_EVENTS_NAME;              // String
		public static final String EVENT_TYPE_KEY = PriceFxInterface.CHARTDEF_EVENTS_TYPE;              // String
		public static final String EVENT_SUBTYPE_KEY = PriceFxInterface.CHARTDEF_EVENTS_SUBTYPE;        // String

		private List<TriggerEventImpl> triggeredEvents = new ArrayList<>();

		public class TriggerEventImpl implements TriggerEvent {

			private final String eventName;

			private final List<String> pointAttributes = new ArrayList<>();

			TriggerEventImpl(final String eventName) {
				this.eventName = eventName;
			}

			public TriggerEvent withXAsEventDataAttr() {
				pointAttributes.add("x");
				return this;
			}

			public TriggerEvent withYAsEventDataAttr() {
				pointAttributes.add("y");
				return this;
			}

			public TriggerEvent withNameAsEventDataAttr() {
				pointAttributes.add("name");
				return this;
			}

			public TriggerEvent withCategoryAsEventDataAttr() {
				pointAttributes.add("category");
				return this;
			}

			public TriggerEvent withCustomEventDataAttr(final String customAttribute) {
				if (customAttribute == null || customAttribute.isEmpty()) {
					throw new IllegalArgumentException("TriggerEvent.withCustomEventDataAttr(customAttribute) <customAttribute> should be neither null nor empty");
				}
				pointAttributes.add(customAttribute);
				return this;
			}

			public OnPointClick back() {
				return OnPointClickImpl.this;
			}

			List<Map<String, String>> getAsMaps() {
				final List<Map<String, String>> list = new ArrayList<>();
				for (String attr : pointAttributes) {
					final Map<String, String> eventMap = new HashMap<>();
					eventMap.put(EVENT_TYPE_KEY, "onClick");
					eventMap.put(EVENT_NAME_KEY, eventName);
					eventMap.put(EVENT_SUBTYPE_KEY, attr);
					list.add(eventMap);
				}
				return list;
			}

		}


		public TriggerEvent triggerEvent(final String eventName) {
			if (eventName == null || eventName.isEmpty()) {
				throw new IllegalArgumentException("OnPointClick.triggerEvent(eventName) <eventName> should be neither null nor empty");
			}
			final TriggerEventImpl event = new TriggerEventImpl(eventName);
			triggeredEvents.add(event);
			return event;
		}

		public ResultHighchart back() {
			return ResultHighchart.this;
		}

		List<Map<String, String>> getEvents() {
			return triggeredEvents.stream().map(TriggerEventImpl::getAsMaps).flatMap(List::stream).collect(Collectors.toList());
		}


	}


	/**
	 * <b>BETA</b>
	 * <br/>
	 * Allows to define events that will be triggered when the user clicks in the chart/map.
	 * <br>
	 * Example:
	 * <pre>
	 *     def chart = api.buildHighchart(definition)
	 *     chart.onPointClick().triggerEvent(api.dashboardWideEvent("eventName"))
	 *         .withNameAsEventDataAttr().withCustomEventDataAttr("value")
	 *     return chart
	 * </pre>
	 * This allows to refresh an embedded dashboard:
	 * <pre>
	 *     return api.dashboard("embeddedDashboardName")
	 *         //.setParam("country_name", api.input("Country"))
	 *         //.setParam("country_revenue", api.input("Revenu"))
	 *         .showEmbedded()
	 *         .andRecalculateOn(api.dashboardWideEvent("eventName"))
	 *         .withEventDataAttr("name").asParam("country_name)
	 *         .withEventDataAttr("value").asParam("country_revenue")
	 * </pre>
	 */
	public OnPointClick onPointClick() {
		if (onClick == null) {
			onClick = new OnPointClickImpl();
		}
		return onClick;
	}

	/**
	 * Not intended for public use, required for JSON serialization.
	 */
	public List<Map<String, String>> getEvents() {
		final List<Map<String, String>> events = new ArrayList<>();
		if (onClick != null) {
			events.addAll(onClick.getEvents());
		}
		return events;
	}

}
